/*

  Ray 1.0 - Billboarding maps

*/


#ifndef _LIB_BILLBOARD_MAP
#define _LIB_BILLBOARD_MAP


/* discrete 2-D billboarding points 
   for spherical integral approximations  */

typedef struct {
 FTYPE *start_x; // array of starting x-coordinates
 int *steps;     // number of scan-line horizontal steps
 int height;      // total number of scan-lines
 FTYPE start_y;   // starting y coordinate
 int area;        // total number of 2-D samples
 FTYPE inv_area;  // inverse of total number of 2-D samples
 FTYPE delta;     // delta between rows - scan-lines (same for x and y)
 t_basis3d b;     // billboarding basis
} t_billboard_map;

/* TODO: for now only circular */
void make_billboard_map(t_billboard_map *map, float radius, int subdiv)
{
int j;
FTYPE x, y, delta;

if(subdiv < 1) subdiv = 1;

map->start_x = (FTYPE *)malloc( subdiv * sizeof(FTYPE) );
map->start_y = -radius;
map->steps   = (int *)malloc( subdiv * sizeof(int) );
map->delta = delta = radius*2.0 / (FTYPE)subdiv;
map->height = subdiv;

y = -radius;
radius*=radius;

 for(j=0; j < subdiv; j++, y+=delta)
 {
  x = SQRT( radius - y*y );
  map->start_x[j] = -x;
  map->steps[j] = (int)(x*2.0 / delta);
	map->area += map->steps[j];
 }

map->inv_area = 1.0 / (FTYPE)map->area;

}


void orient_billboard_map(t_billboard_map *map, t_vector3d *fwd)
{
 static t_vector3d up = { 0.0, 1.0, 0.0 }; /* TODO */
 t_basis3d *b = &map->b;

 prod_v( &b->u, fwd, &up );
 prod_v( &b->v, &b->u, fwd );
  norm_v( &map->b.u );
  norm_v( &map->b.v );
 prod_v( &b->w, &b->v, &b->u );
  //norm_v( &map->b.w );
}



#endif

