/*

 Ray 1.0 - (8-8-8)RGB colors handling library

*/



#ifndef _LIB_DRAW24
#define _LIB_DRAW24

// packs 3 integers ranging from 0 to 255 into a 24bit int value
#define RGB2INT(r,g,b)  (((int)(r)&0xFF)<<16)|(((int)(g)&0xFF)<<8)|((int)(b)&0xFF)

// packs 3 FTYPEs ranging from 0.0 to 1.1 into a 24bit int value
#define RGBF2INT(r,g,b) RGB2INT(r*255.0f, g*255.0f, b*255.0f)


// current drawing color
__int32 g_MyDrawColor32 = 0x000000; // Black by default

// linear frame buffer pointer
__int32 *g_MySurface = NULL;








inline __int32 blend_to_white(__int32 c, __int32 val)
{
__int32 val1;
// val&=0xFF;
val1 = 255-val;
 return (
	 ( ( ((c & 0xFF00FF) * val1) + (0xFF00FF * val) )>>8)&0xFF00FF | 
	 ( ( ((c & 0x00FF00) * val1) + (0x00FF00 * val) )>>8)&0x00FF00
	 );
}

inline __int32 blend_to_black(__int32 c, __int32 val)
{
//val&=0xFF;
val = 255-val;
 return (
	 (( ((c & 0xFF00FF)*val) )>>8)&0xFF00FF | 
	 (( ((c & 0x00FF00)*val) )>>8)&0x00FF00
	 );
}

inline __int32 blend_to_color(__int32 c1, __int32 c2, __int32 val)
{
__int32 val1;
//val &= 0xFF;
val1 = 255 - val;
 return (
	 (( ((c1 & 0xFF00FF) * val1) + 
	    ((c2 & 0xFF00FF) * val )  )>>8) & 0xFF00FF | 
	 (( ((c1 & 0x00FF00) * val1) + 
	    ((c2 & 0x00FF00) * val )  )>>8) & 0x00FF00
	 );
}

__inline void put_pixel_32(int x, int y)
{
 if((unsigned int)x > SCREEN_X-1) return;
 if((unsigned int)y > SCREEN_Y-1) return;
 *(g_MySurface + x + y*SCREEN_X) = g_MyDrawColor32;
}


__inline void put_pixel_AA32(FTYPE x, FTYPE y)
{
if((unsigned int)x > SCREEN_X-2) return;
if((unsigned int)y > SCREEN_Y-2) return;

//__int32 col = g_MyDrawColor32;
__int32 *buf = g_MySurface + ((int)(x) + ((int)(y))*SCREEN_X);
int b, d, t;

b = (int)(x*256.0) & 0xFF;
d = (int)(y*256.0) & 0xFF;

t = (b * d) >> 8;

 *(buf)       = blend_to_color(*buf,         g_MyDrawColor32,  256 - b - d + t );
 *(buf+1)     = blend_to_color(*(buf+1),     g_MyDrawColor32,  b - t );
 *(buf+SCREEN_X)   = blend_to_color(*(buf+SCREEN_X),   g_MyDrawColor32,  d - t );
 *(buf+SCREEN_X+1) = blend_to_color(*(buf+SCREEN_X+1), g_MyDrawColor32,  t );

}





FTYPE Q_rsqrt( FTYPE number )
{
long i;
FTYPE x2, y;
const FTYPE threehalfs = 1.5F;

x2 = number * 0.5F;
y = number;

i = * ( long * ) &y; // evil floating point bit level hacking
i = 0x5f3759df - ( i >> 1 ); // what !?
y = * ( FTYPE * ) &i;

y = y * ( threehalfs - ( x2 * y * y ) ); // 1st iteration
//y = y * ( threehalfs - ( x2 * y * y ) ); // 2nd iteration, this can be removed

return y;
}


void draw_rect(int x1, int y1, int x2, int y2)
{
int t, l, h;
__int32 *v, *v1, *v2;

if( ((unsigned int)x1 >= SCREEN_X) || ((unsigned int)y2 >= SCREEN_Y)  ) return;
if( ((unsigned int)y1 >= SCREEN_Y) || ((unsigned int)y2 >= SCREEN_Y)  ) return;

if(x1 > x2) { t=x1; x1=x2; x2=t; }
if(y1 > y2) { t=y1; y1=y2; y2=t; }

l = x2 - x1;
h = y2 - y1;

v = g_MySurface  +  x1 + y1*SCREEN_X;

v1 = v;
v2 = v + h*SCREEN_X;

for(t=0; t<=l; t++)
{
 *v1 = *v2 = g_MyDrawColor32;
 v1++; v2++;
}

v1 = v + SCREEN_X;
v2 = v + l;
for(t=1; t<=h; t++)
{
 *v1 = *v2 = g_MyDrawColor32;
 v1+=SCREEN_X; 
 v2+=SCREEN_X;
}

}



void fill_rect(int x1, int y1, int x2, int y2, int alpha)
{
int t;

if( ((unsigned int)x1 >= SCREEN_X) || ((unsigned int)y2 >= SCREEN_Y)  ) return;
if( ((unsigned int)y1 >= SCREEN_Y) || ((unsigned int)y2 >= SCREEN_Y)  ) return;

if(x1 > x2) { t=x1; x1=x2; x2=t; }
if(y1 > y2) { t=y1; y1=y2; y2=t; }

int l = x2 - x1;
int h = y2 - y1;

__int32 *v = g_MySurface  +  x1 + y1*SCREEN_X;

alpha &= 0xFF;

int i,j;
int voff = SCREEN_X - l;

for(j=0; j < h; j++)
{
  for(i=0; i < l; i++)
  {
     *v = blend_to_color( *v, g_MyDrawColor32, alpha);
	 v++;
  }
  v += voff;
}


}




/*
void drawLineAA32(FTYPE x1, FTYPE y1, FTYPE x2, FTYPE y2)
{
FTYPE sx, sy, t;

// quick (temporary) clipping
if((unsigned int)(x1-1) > SDL_X_RES-4) return;
if((unsigned int)(y1-1) > SDL_Y_RES-4) return;
if((unsigned int)(x2-1) > SDL_X_RES-4) return;
if((unsigned int)(y2-1) > SDL_Y_RES-4) return;

 // calculate length of segment
 sx = x1-x2;
  t = sx*sx;
 sx = y2-y1;
  //t = sqrtf(t + sx*sx); // the only expensive op...

 t = Q_rsqrt(t + sx*sx);

 if(t > 256.0)  putPixelAA32((x1+x2)*.5, (y1+y2)*.5);

 // if the line is too short draw a pixel instead
 // (anti-aliasing color depth resolution is 1/256 for each color component)
 // if(t < 1.0/256.0)  putPixelAA32((x1+x2)*.5, (y1+y2)*.5);

  else

  {
   //t+=1.0;
   int i = 1.0/t + .5;
   __int32 *buf;
   __int32 b, d, k;
   __int32 px, py, dx, dy;
   __int32 c1, c2a, c2b, val, val1;


   px = x1 * (FTYPE)((1<<RGB_FIXED_POINT_2_F) - 1);
   py = y1 * (FTYPE)((1<<RGB_FIXED_POINT_2_F) - 1);

   // calculate delta-x and delta-y as Delta / Distance
   
   t*=((1<<RGB_FIXED_POINT_2_F) - 1);

   dx = (x2 - x1) * t;
   dy = (y2 - y1) * t;


//   g_MyDrawColor32 = 0xFFFFFF;
   c2a = g_MyDrawColor32 & 0xFF00FF;
   c2b = g_MyDrawColor32 & 0x00FF00;


    for(; i; i--, px+=dx, py+=dy)//  putPixelAA32(sx+=dx, sy+=dy);  
	{

     buf = g_MySurface + ( (px>>RGB_FIXED_POINT_2_F) + (py>>RGB_FIXED_POINT_2_F)*640 );
	

     b = (px>>(RGB_FIXED_POINT_2_F - 8)) & 0xFF;
     d = (py>>(RGB_FIXED_POINT_2_F - 8)) & 0xFF;
     k = (b * d) >> 8;

// *buf = blend_to_Color32(*buf,  g_MyDrawColor32,  256 - b - d + k );

 c1 = *buf; val = 256 - b - d + k;

val1 = 255 - val;
*buf =
     ((( ((c1 & 0xFF00FF) * val1) + 
	    ((c2a) * val )  )>>8) & 0xFF00FF) | 
	 ((( ((c1 & 0x00FF00) * val1) + 
	    ((c2b) * val )  )>>8) & 0x00FF00) ;

buf++;


// *buf = blend_to_Color32(*buf, g_MyDrawColor32,  b - k );
 c1 = *buf; val = b - k;

val1 = 255 - val;
*buf =
     ((( ((c1 & 0xFF00FF) * val1) + 
	    ((c2a) * val )  )>>8) & 0xFF00FF) | 
	 ((( ((c1 & 0x00FF00) * val1) + 
	    ((c2b) * val )  )>>8) & 0x00FF00) ;

buf+=640-1;

// *buf = blend_to_Color32(*buf,   g_MyDrawColor32,  d - k );

 c1 = *buf; val = d - k;

val1 = 255 - val;
*buf =
     ((( ((c1 & 0xFF00FF) * val1) + 
	    ((c2a) * val )  )>>8) & 0xFF00FF) | 
	 ((( ((c1 & 0x00FF00) * val1) + 
	    ((c2b) * val )  )>>8) & 0x00FF00) ;




buf++;

//*buf = blend_to_Color32(*buf, g_MyDrawColor32,  k );
 c1 = *buf; val = k;

val1 = 255 - val;
*buf =
     ((( ((c1 & 0xFF00FF) * val1) + 
	    ((c2a) * val )  )>>8) & 0xFF00FF) | 
	 ((( ((c1 & 0x00FF00) * val1) + 
	    ((c2b) * val )  )>>8) & 0x00FF00) ;




	}
	
	

  }

}
*/

#endif

