/*

  Ray 1.0 - Fixed Point Math library

*/
#ifndef _LIB_FIXED
#define _LIB_FIXED


// Define fixed point integer type (32 bits integers):
#define FIX_TYPE __int32

// How much bits are used for the fractional part
#define FIX_BITS 16

// Macros used to pass from float values
// to fixed-point values and vice-versa:
#define FLOAT_2_FIX( val )     ((FIX_TYPE)( (FTYPE)(val) * (FTYPE)((1L << FIX_BITS)-1) ))
#define FLOAT_2_FIX256( val )  ((FIX_TYPE)( (FTYPE)(val) * (FTYPE)((1L << (FIX_BITS + 8))-1) ))
#define FIX_2_FLOAT( val )     ((FTYPE)(val) * (1.0/(FTYPE)((1L << FIX_BITS)-1)) )
#define FIX_2_BYTE( val )      ( ((FIX_TYPE)(val) >> (FIX_BITS - 8)) )
#define FIX_2_INT( val )       ( (FIX_TYPE)(val) >> FIX_BITS ) 



/*
Quad topology:

 p0--p1
 |    |
 p2--p3
*/

#define PIXEL_NUM_VARS 8
 typedef struct {
 FIX_TYPE values[PIXEL_NUM_VARS];
/*
 FIX_TYPE dr, dg, db; // diffuse components
 FIX_TYPE tx, ty, tz; // texture coordinates
 FIX_TYPE sr, sg, sb; // specular components
*/
 int shape_id; // ID of the shape wich generated the pixel
 void *mat;
 __int32 *texture;
} t_pixel;




#endif
