/*

 Ray 1.0 - Materaials handlig library

*/

#ifndef _LIB_MATERIALS
#define _LIB_MATERIALS


#include "colors.h"
#include "noise.h"



#define DEFAULT_SHINYNESS 90.0


typedef struct {

	t_texture_enum tex;  // type of texture

	__int32 *texture;

//  bool double_sided; // is material double sided ?

  t_rgb diffuse;  // material diffuse rgb color

  t_rgb specular; // material specular rgb color

  t_rgb emission; // material emissive rgb color

  FTYPE shinyness;  // material shinyness: sh.= -(1/sh) + (1+sh)/(sh*(cos_beta^2+1))
   FTYPE shinyness_inv; // (offset constant for shinyness computations)
  FTYPE transparency; // material transparency [0.0 - 1.0]
  FTYPE reflectance;  // material reflectance  [0.0 - 1.0]
  FTYPE ior;  // material index of refraction

  //t_rgba (*texture_fun)(FTYPE*); // diffuse texturing function pointer

} t_material;


/*

 Texture mapping functions :

*/
#ifndef LINUX_COMPILER  
// Borland and Microsoft

 t_rgba __cdecl texture2d(FTYPE*);

 t_rgba __cdecl texture3d(FTYPE*);

#else // GNU GCC

 t_rgba texture2d(FTYPE*) __attribute__((cdecl));

 t_rgba texture3d(FTYPE*) __attribute__((cdecl));

#endif






#define MAX_MATERIALS 128

t_material *g_materials[MAX_MATERIALS];
t_material *g_cur_material = NULL;
int g_num_materials = 0;





// XOR3D texture
#define XOR3D(x,y,z)  ((((int)(x*.5*255.0)^(int)(y*.5*255.0)^(int)(z*.5*255.0))&0xFF)/255.0)

#define CHECKER3D(x,y,z) ( ((unsigned char)(4.0*(x+2.001)) + (unsigned char)(4.0*(y+2.001)) + (unsigned char)(4.0*(z+2.001))) &1)





// rainbow palette
static t_rgb g_rainbow[7] = {
	{0,1,1}, // cyan
	{0,0,1}, // blue
	{1,0,1}, // violet
	{1,0,0}, // red
	{1,1,0}, // yellow
	{0,1,0}, // green

	{0,1,1}  // wrap - cyan
};


void set_material_shinyness(t_material *m, FTYPE sh)
{
 m->shinyness = fabs( sh );
 if( m->shinyness ) m->shinyness_inv = 1.0 / m->shinyness;
}


void set_material_default(t_material *m)
{
 set_color3f( &m->diffuse,  1,1,1 );
 set_color3f( &m->emission, 0,0,0 );
 set_color3f( &m->specular, 1,1,1 );
 set_material_shinyness( m, DEFAULT_SHINYNESS );
 m->reflectance = .6;
// m->double_sided = false;
}


t_material *allocate_material()
{
t_material *m;

  if(g_num_materials >= MAX_MATERIALS ) return NULL;

   m = (t_material *)malloc( sizeof(t_material) );

  if( m == NULL ) return NULL;

// initializes default common values:
 set_material_default( m );

 g_cur_material = g_materials[ g_num_materials ] = m;

 g_num_materials++;

 return m;
}


t_material *make_material( t_material_enum type )
{
/* todo */
 return allocate_material();
}


void color_rainbow(FTYPE *dr, FTYPE *dg, FTYPE *db, FTYPE s)
{
FTYPE s1;
FTYPE r, g, b;
FTYPE r1, g1, b1;
t_rgb *c;

//if(!dest) return;

if(s < 0.0) s = -s;
if(s >= 1.0) s = fmod(s, 1.0);
s*=6.0;

 c = &g_rainbow[(int)s];

 s  = fmod(s, 1.0);
 s1 = 1.0 - s;

 r1  = c->r;
 g1  = c->g;
 b1  = c->b;

  c++;
 r = c->r;
 g = c->g;
 b = c->b;


 *dr = r * s  +  r1 * s1;
 *dg = g * s  +  g1 * s1;
 *db = b * s  +  b1 * s1;
                              
}





#endif

