/* 
 Ray 1.0
  
 math3d.h

 Procedures to create, handle and use 3d vectors and 3d-basis:
*/

#ifndef _LIB_MATH3D
#define _LIB_MATH3D

typedef struct { FTYPE x, y, z; } t_point3d;
typedef struct { FTYPE x, y, z; } t_vector3d;

typedef struct { FTYPE x, y, z, w; } t_point4d;
typedef struct { FTYPE x, y, z, w; } t_vector4d;

typedef struct { t_point3d o; t_vector3d u, v, w; } t_basis3d;



typedef struct {
 t_point3d  o; // origin of the ray
 t_vector3d d; // direction of the ray
} t_ray;


/* Vector-handling functions: */

/* Basis-handling functions: */

/* Misc. vector-basis functions: */


void print_b(t_basis3d *b)
{
#ifdef DEBUG_LOG
printf("BASIS PRINT:\no(%.2f,%.2f,%.2f)\nu(%.2f,%.2f,%.2f)\nv(%.2f,%.2f,%.2f)\nw(%.2f,%.2f,%.2f)\n---\n",
 b->o.x,b->o.y,b->o.z,
 b->u.x,b->u.y,b->u.z,
 b->v.x,b->v.y,b->v.z,
 b->w.x,b->w.y,b->w.z );
#endif
}


////////////////////////////////////
// VECTOR-FUNCTIONS DEFINITIONS:  //
////////////////////////////////////

void create_v(t_point3d * start, t_point3d * end, t_vector3d * res)
{
 res->x = end->x - start->x;
 res->y = end->y - start->y;
 res->z = end->z - start->z;
}

void add_p_v(t_point3d *p, t_vector3d *v)
{
 p->x += v->x;
 p->y += v->y;
 p->z += v->z;
}

void add_v(t_vector3d *v1, t_vector3d *v2)
{
 v1->x += v2->x;
 v1->y += v2->y;
 v1->z += v2->z;
}

void sub_p_v(t_point3d *p, t_vector3d *v)
{
 p->x -= v->x;
 p->y -= v->y;
 p->z -= v->z;
}

void sub_v(t_vector3d *v1, t_vector3d *v2)
{
 v1->x -= v2->x;
 v1->y -= v2->y;
 v1->z -= v2->z;
}

FTYPE prod_v_p(t_vector3d *v, t_point3d *p)
{
 return v->x*p->x + v->y*p->y + v->z*p->z; 
}


void prod_v_k(t_vector3d *v, FTYPE k)
{
 v->x*=k;
 v->y*=k;
 v->z*=k;
}

FTYPE length_v(t_vector3d * u)
{
 return SQRT(u->x * u->x + u->y * u->y + u->z * u->z);
}

FTYPE prod_scalar (t_vector3d * u, t_vector3d * v)
{
 return (u->x * v->x + u->y * v->y + u->z * v->z);
}


void norm_v(t_vector3d *v)
{
FTYPE len;

 len = length_v( v );

 if(len)
 {
  len = 1.0f/len;
  v->x *= len;
  v->y *= len;
  v->z *= len;
 }
}


void norm_v_k(t_vector3d *v, FTYPE k)
{
FTYPE len;
 len = length_v( v );
 if(len)
 {
  len = k / len;
  v->x *= len;
  v->y *= len;
  v->z *= len;
 }
}

// Z coordinates are negated to work in the reference system
// (xy) is the screen plane
//  y ^
//    | / z
//    |/
//    O-->x
//   
void prod_v(t_vector3d *res, t_vector3d *u, t_vector3d *v)
{
  res->x = (u->z * v->y) - (u->y * v->z);
  res->y = (u->x * v->z) - (u->z * v->x);
  res->z = (u->y * v->x) - (u->x * v->y);
}


///////////////////////////////////
// BASIS-FUNCTIONS DEFINITIONS:  //
///////////////////////////////////

void create_basis(t_basis3d *b, t_vector3d *forward, t_vector3d *up)
{
 prod_v( &b->u, forward, up);
 prod_v( &b->v, &b->u, forward);
 prod_v( &b->w, &b->v, &b->u);
}


void transform_p(t_point3d *d, t_point3d *s, t_basis3d *b)
{
  d->x  =  s->x * b->u.x  +  s->y * b->v.x  +  s->z * b->w.x   +   b->o.x;
  d->y  =  s->x * b->u.y  +  s->y * b->v.y  +  s->z * b->w.y   +   b->o.y;
  d->z  =  s->x * b->u.z  +  s->y * b->v.z  +  s->z * b->w.z   +   b->o.z;
}

void transform_v(t_vector3d *d, t_vector3d *s, t_basis3d *b)
{
  d->x  =  s->x * b->u.x  +  s->y * b->v.x  +  s->z * b->w.x;
  d->y  =  s->x * b->u.y  +  s->y * b->v.y  +  s->z * b->w.y;
  d->z  =  s->x * b->u.z  +  s->y * b->v.z  +  s->z * b->w.z;
}



void norm_b(t_basis3d *b)
{
 norm_v( &b->u );
 norm_v( &b->v );
 norm_v( &b->w ); 
}

void norm_b_k(t_basis3d *b, FTYPE k)
{
 norm_v_k( &b->u, k );
 norm_v_k( &b->v, k );
 norm_v_k( &b->w, k ); 
}

void norm_b_3k(t_basis3d *b, FTYPE k1, FTYPE k2, FTYPE k3 )
{
 norm_v_k( &b->u, k1 );
 norm_v_k( &b->v, k2 );
 norm_v_k( &b->w, k3 ); 
}

void create_norm_b(t_basis3d *b, t_vector3d *forward, t_vector3d *up)
{
 prod_v( &b->u, forward, up );
 prod_v( &b->v, &b->u, forward );
 //prod_v( &b->w, &b->v, &b->u );
	b->w = *forward;

 norm_b( b );
}

void create_norm_b_k(t_basis3d *b, t_vector3d *forward, t_vector3d *up, FTYPE k)
{
 prod_v( &b->u, forward, up);
 prod_v( &b->v, &b->u, forward);
 //prod_v( &b->w, &b->v, &b->u);
 b->w = *forward;

 norm_b_k( b, k );
}

void create_norm_b_3k(t_basis3d *b, t_vector3d *forward, t_vector3d *up, FTYPE k1, FTYPE k2, FTYPE k3 )
{
 prod_v( &b->u, forward, up);
 prod_v( &b->v, &b->u, forward);
 //prod_v( &b->w, &b->v, &b->u);
  b->w = *forward;

 norm_b_3k( b,  k1, k2, k3 );
}


/* Misc. functions */

void average_v(t_vector3d *dst, t_vector3d *v1, t_vector3d *v2)
{
 dst->x = ( v1->x + v2->x ) * .5;
 dst->y = ( v1->y + v2->y ) * .5;
 dst->z = ( v1->z + v2->z ) * .5;
}


void interpolate_v(t_vector3d *v1, t_vector3d *v2, FTYPE k, t_vector3d *res)
{
FTYPE k1 = 1.0f - k;

 res->x = v1->x * k1  +  v2->x * k;
 res->y = v1->y * k1  +  v2->y * k;
 res->z = v1->z * k1  +  v2->z * k;

 //norm_v(res);
}



#endif

