/* 

 Ray 1.0 - basic 4x4 Matrix library

 - Original version by Ken Perlin

*/


#ifndef _LIB_MATRIX3D
#define _LIB_MATRIX3D


#include <math.h>

#ifdef LINUX_COMPILER
#include <string.h>
#endif


typedef FTYPE t_matrix[4*4]; // Matrix as a 4x4 array of FTYPEs


t_matrix g_M_S[128]; // Matrix Stack
int g_M_SP = 0;      // Matrix Stack-Pointer
t_matrix g_M_tmp;    // temporary matrix




void m_print(FTYPE *m)
{
int i, j;


			for(i=0; i<4; i++)
			{
				for(j=0; j<4; j++)
				{
					printf( " %.2f", m[i*4+j] );
				}
				printf("\n");
			}
}




   void m_identity(FTYPE *matrix) {
      for (int i = 0 ; i < 4 ; i++)
      for (int j = 0 ; j < 4 ; j++)
         matrix[i*4+j] = (i == j ? 1 : 0);
   }

   void m_identity_k(FTYPE *matrix, FTYPE k) {
      for (int i = 0 ; i < 4 ; i++)
      for (int j = 0 ; j < 4 ; j++)
         matrix[i*4+j] = (i == j ? k : 0);
   }



   void m_copy(FTYPE *src, FTYPE *dst) {
    memcpy(dst, src, 4*4*sizeof(FTYPE));
   }


   void m_preMultiply(FTYPE *dst, FTYPE *b)
	 {
	  static t_matrix tmp;
		int off;

      m_copy(dst, tmp);

      for (int i = 0 ; i < 4 ; i++)
      for (int j = 0 ; j < 4 ; j++)
			{
         dst[i*4+j] = 0;
         for (int k = 0 ; k < 4 ; k++)
            dst[i*4+j] += tmp[i*4+k] * b[k*4+j];
      }


   }


   void m_postMultiply(FTYPE *dst, FTYPE *b)
   {
		static t_matrix tmp;

      m_copy(dst, tmp);

      for (int i = 0 ; i < 4 ; i++)
       for (int j = 0 ; j < 4 ; j++)
	   {
         dst[i*4+j] = 0;

         for (int k = 0 ; k < 4 ; k++)

            dst[i*4+j] += b[i*4+k] * tmp[k*4+j];
       }
   }



//----- ROUTINES TO GENERATE TRANSFORMATION MATRICES -----

   void m_Translation(FTYPE *m, FTYPE x, FTYPE y, FTYPE z)
   {
      m_identity(m);

      m[0*4+3] = x;
      m[1*4+3] = y;
      m[2*4+3] = z;
   }

   void m_Rotation(FTYPE *m, int i, int j,FTYPE theta)
   {
      m_identity(m);

      m[i*4+i] = m[j*4+j] = cosf(theta);
      m[i*4+j] = -sinf(theta);
      m[j*4+i] = -m[i*4+j];
   }

   void m_Scale(FTYPE *m, FTYPE x,FTYPE y,FTYPE z)
   {
      m_identity(m);

      m[0*4+0] = x;
      m[1*4+1] = y;
      m[2*4+2] = z;
   }


   //----- ROUTINE TO TRANSFORM A POINT -----

   void m_transform_point(FTYPE *m, t_point3d *p, t_point3d *dst)
   {
      //for (int j = 0 ; j < 3 ; j++)
        // dst[j] = m[j*4+0]*p->x + m[j*4+1]*p->y + m[j*4+2]*p->z + m[j*4+3];
		  dst->x = m[0*4+0]*p->x + m[0*4+1]*p->y + m[0*4+2]*p->z + m[0*4+3];
			dst->y = m[1*4+0]*p->x + m[1*4+1]*p->y + m[1*4+2]*p->z + m[1*4+3];
			dst->z = m[2*4+0]*p->x + m[2*4+1]*p->y + m[2*4+2]*p->z + m[2*4+3];
   }

   void m_transform_vector(FTYPE *m, t_vector3d *p, t_vector3d *dst)
   {
      //for (int j = 0 ; j < 3 ; j++)
        // dst[j] = m[j*4+0]*p->x + m[j*4+1]*p->y + m[j*4+2]*p->z + m[j*4+3];
		  dst->x = m[0*4+0]*p->x + m[0*4+1]*p->y + m[0*4+2]*p->z;
			dst->y = m[1*4+0]*p->x + m[1*4+1]*p->y + m[1*4+2]*p->z;
			dst->z = m[2*4+0]*p->x + m[2*4+1]*p->y + m[2*4+2]*p->z;
   }

   void m_transform_normal(FTYPE *m_inv, t_vector3d *p, t_vector3d *dst)
   {
      //for (int j = 0 ; j < 3 ; j++)
        // dst[j] = m[j*4+0]*p->x + m[j*4+1]*p->y + m[j*4+2]*p->z + m[j*4+3];
		  dst->x = m_inv[0*4+0]*p->x + m_inv[1*4+0]*p->y + m_inv[2*4+0]*p->z;
			dst->y = m_inv[0*4+1]*p->x + m_inv[1*4+1]*p->y + m_inv[2*4+1]*p->z;
			dst->z = m_inv[0*4+2]*p->x + m_inv[1*4+2]*p->y + m_inv[2*4+2]*p->z;
   }

   void m_transform_ray(FTYPE *m, t_ray *r, t_ray *dst)
	 {
		 // transform the origin
      m_transform_point( m, &r->o, &dst->o );
		 // transform the direction
      m_transform_vector( m, &r->d, &dst->d );
	 }




   void init_m() {
      g_M_SP = 0;
      m_identity(g_M_S[g_M_SP]);
   }

   void m_push() {
      m_copy(g_M_S[g_M_SP], g_M_S[g_M_SP+1]);
      g_M_SP++;
   }

   void m_pop() {
      --g_M_SP;
   }






//----- ROUTINES TO ROTATE AND TRANSLATE MATRICES -----

   void m_translate(FTYPE *m, FTYPE tx, FTYPE ty, FTYPE tz)
	 {
		 static t_matrix tmp;
      m_Translation(tmp, tx, ty, tz);
      m_preMultiply(m, tmp);
   }

   void m_rotate_x(FTYPE *m, FTYPE theta) 
	 {
		 static t_matrix tmp;
      m_Rotation(tmp, 1,2, theta);
      m_preMultiply(m, tmp);
   }

   void m_rotate_y(FTYPE *m, FTYPE theta) 
	 {
		 static t_matrix tmp;
      m_Rotation(tmp, 2,0, theta);
      m_preMultiply(m, tmp);
   }

   void m_rotate_z(FTYPE *m, FTYPE theta) 
	 {
		 static t_matrix tmp;
      m_Rotation(tmp, 0,1, theta);
      m_preMultiply(m, tmp);
   }



/*
   void m_rotate(FTYPE *m, FTYPE theta, FTYPE dx, FTYPE dy, FTYPE dz) 
	 {
		 static t_matrix tmp;

      m_preMultiply(m, tmp);
   }
*/



   void m_scale(FTYPE *m, FTYPE sx, FTYPE sy, FTYPE sz) 
	 {	 
		 static t_matrix tmp;
      m_Scale(tmp, sx, sy, sz);
      m_preMultiply(m, tmp);
   }


//----- INVERTING A 4x4 THAT WAS CREATED BY TRANSLATIONS+ROTATIONS+SCALES

   void m_invert(FTYPE *src, FTYPE *dst) {
   int i, j;

      // COMPUTE ADJOINT COFACTOR MATRIX FOR THE ROTATION+SCALE 3x3

      for (i = 0 ; i < 3 ; i++)
      for (j = 0 ; j < 3 ; j++) {
         int i0 = (i+1) % 3;
         int i1 = (i+2) % 3;
         int j0 = (j+1) % 3;
         int j1 = (j+2) % 3;
         dst[j*4+i] = src[i0*4+j0] * src[i1*4+j1] - src[i0*4+j1] * src[i1*4+j0];
      }

      // RENORMALIZE BY DETERMINANT TO GET ROTATION+SCALE 3x3 INVERSE

      FTYPE determinant =  src[0*4+0] * dst[0*4+0] +
                           src[1*4+0] * dst[0*4+1] +
                           src[2*4+0] * dst[0*4+2];

			
      for (i = 0 ; i < 3 ; i++)
      for (j = 0 ; j < 3 ; j++)
         dst[i*4+j] /= determinant;

      // COMPUTE INVERSE TRANSLATION
      for (i = 0 ; i < 3 ; i++)
         dst[i*4+3]= - dst[i*4+0] * src[0*4+3]
                     - dst[i*4+1] * src[1*4+3]
                     - dst[i*4+2] * src[2*4+3] ;


#ifdef DEBUG_LOG
 printf("Inverse matrix computed:\n");
 m_print( src );
#endif
   

}







#endif

