#ifndef _LIBRERIA_OGGETTI3D_
#define _LIBRERIA_OGGETTI3D_


#include "Motore3D.h"
#include "Bezier.h"


// Valori che identificano univocamente il tipo di oggetto
#define M3D_GRIGLIA      0
#define M3D_CUBO         1 
#define M3D_CONO         2
#define M3D_CILINDRO     3
#define M3D_TOROIDE      4
#define M3D_SFERA        5
#define M3D_TUBO         6
#define M3D_TUBO_PTS     7



//#define M3D_GEN_LOG_OBJ_INFO



#ifdef M3D_GEN_LOG_OBJ_INFO
char M3D_obj_base_name[][] =
  {
   "griglia quadrata: dim(l %.2f, h %.2f),   num(x %i, y %i)\n",
   "cubo:             dim(l %.2f, h %.2f, p %.2f), num(x %i, y %i, z %i)\n",
   "cono (lathe):     dim(r %.2f, h %.2f),   num(rad %i, h %i)\n",
   "cilindro (lathe): dim(r %.2f, h %.2f),   num(rad %i, h %i)\n",
   "toroide (lathe):  dim(r %.2f, h %.2f),   num(rad %i, h %i)\n",
   "sfera (lathe):    dim(r %.2f, h %.2f),   num(rad %i, h %i)\n",
   "tubazione:        prm(P %.2f, Q%.2f, R%.2f),  num(nl %i, nr %i)\n",
   "tubazione(spline):num(nl %i, nr %i)\n"
  };
#endif


// PROTOTIPI:

 // ROUTINES PER CREARE VARI TIPI DI OGGETTI:
Oggetto3d* crea_griglia(FTYPE l, FTYPE h, int _NSL, int _NSH, FTYPE wx, FTYPE wy);
Oggetto3d* crea_cubo(FTYPE l );
Oggetto3d* crea_cono(FTYPE r, FTYPE h, int _NSB);
Oggetto3d* crea_cilindro(FTYPE h, FTYPE r, int _NSC);
Oggetto3d* crea_toroide( FTYPE r1, FTYPE r2, int _NSR, int _NSC, FTYPE n1, FTYPE n2);
Oggetto3d* crea_sfera( FTYPE r, int _NSH, int _NSC);
Oggetto3d* crea_tubo(FTYPE P, FTYPE Q, FTYPE R,
                FTYPE r1, FTYPE r2, int _NSL, int _NSC,  int nuovo);
Oggetto3d* crea_tubo_pts( Punto3d *pts, int np, Punto3d *shape, int _NSC, FTYPE r2, int nl);

Oggetto3d* crea_tornito(FTYPE *pts, int NP, int _NSC);


//////////////////////
//////////////////////
//////////////////////


void rumore_oggetto( Oggetto3d *src, Oggetto3d *dest, FTYPE ix, FTYPE iy, FTYPE iz, int seed );
void rumore_radiale_oggetto( Oggetto3d *src, Oggetto3d *dest, FTYPE pwr, FTYPE cx, FTYPE cy, FTYPE cz, int seed );
void normalizza_oggetto( Oggetto3d *oggetto, FTYPE dim );



/////////////////
// FUNZIONI:   //
/////////////////

GLvoid glDrawCube()
{
		glBegin(GL_QUADS);
		// Front Face
		glNormal3f( 0.0f, 0.0f, 1.0f);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
		// Back Face
		glNormal3f( 0.0f, 0.0f,-1.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
		glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
		// Top Face
		glNormal3f( 0.0f, 1.0f, 0.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
		// Bottom Face
		glNormal3f( 0.0f,-1.0f, 0.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
		glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
		// Right Face
		glNormal3f( 1.0f, 0.0f, 0.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex3f( 1.0f, -1.0f, -1.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex3f( 1.0f,  1.0f, -1.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex3f( 1.0f,  1.0f,  1.0f);
		glTexCoord2f(0.0f, 0.0f); glVertex3f( 1.0f, -1.0f,  1.0f);
		// Left Face
		glNormal3f(-1.0f, 0.0f, 0.0f);
		glTexCoord2f(0.0f, 0.0f); glVertex3f(-1.0f, -1.0f, -1.0f);
		glTexCoord2f(1.0f, 0.0f); glVertex3f(-1.0f, -1.0f,  1.0f);
		glTexCoord2f(1.0f, 1.0f); glVertex3f(-1.0f,  1.0f,  1.0f);
		glTexCoord2f(0.0f, 1.0f); glVertex3f(-1.0f,  1.0f, -1.0f);
	glEnd();
}



///////////////////////////////////////////////
// ROUTINES PER CREARE VARI TIPI DI OGGETTI: //
///////////////////////////////////////////////

Oggetto3d* crea_cubo_semplice( FTYPE l )
{
int i, iv, ii;
FTYPE l2 = l/2;
FTYPE x, y, z;
FTYPE sqr_vtx[4 * 3] = {-1,  1, -1,
                        -1, -1, -1,
                         1, -1, -1,
                         1,  1, -1};
int cubo_index[3*2 * 6] = {3,1,0, 1,3,2, 7,2,3,
                            2,7,6, 4,6,7, 6,4,5,
                            0,5,4, 5,0,1, 7,0,4,
                            0,7,3, 2,5,1, 5,2,6};
Oggetto3d* oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{
 oggetto = alloca_struct_oggetto();

 oggetto->num_vertici = 8;

 oggetto->num_poly = 12;

 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);
}

iv=0;
for(i=0; i<4*3; i+=3)
{
x = sqr_vtx[i] * l2;
y = sqr_vtx[i+1] * l2;
z = sqr_vtx[i+2] * l2;
 set_localvert_oggetto( oggetto, iv, x, y, z );
 set_localvert_oggetto( oggetto, iv+4, x, y, -z );
 iv++;
}

if(not_present)
{
ii=0;
for(i=0; i<12; i++)
 for(iv=0; iv<3; iv++)
   oggetto->tri[i].v[iv].iv = cubo_index[ii++];


 cuci_vertici_iv( oggetto );

oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}

calcola_normali_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );



return( oggetto );
}








Oggetto3d* crea_cono( FTYPE r, FTYPE h, int _NSB)
{
int ia;
FTYPE a, sa;
FTYPE x, y, z;
int iv;
Oggetto3d *oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{

oggetto = alloca_struct_oggetto();


oggetto->num_poly = _NSB * 2;
oggetto->num_vertici = _NSB + 2;

 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);
}

sa= M_PI*2.0 / (FTYPE)_NSB;
y = -h/2.0;
set_localvert_oggetto( oggetto, 0,  0.0, -y, 0.0 );

iv = 1;

for(ia=0, a=0.0; ia<_NSB; ia++, a+=sa)
{
 x = cos(a) * r;
 z = sin(a) * r;
 set_localvert_oggetto( oggetto, iv, x, y, z );
 iv++;
}
set_localvert_oggetto( oggetto, iv, 0.0, y*1.125, 0);



// Calcola gli indici dei vertici
if(not_present)
{
for(ia=0; ia<_NSB; ia++)
{
// poligoni laterali del cono
 oggetto->tri[ia].v[0].iv = 0;
 oggetto->tri[ia].v[2].iv = ia + 1;
 oggetto->tri[ia].v[1].iv = (ia + 1)%_NSB + 1;

// poligoni di base del cono
 oggetto->tri[ia+_NSB].v[0].iv = oggetto->num_vertici - 1;
 oggetto->tri[ia+_NSB].v[2].iv = (ia + 1)%_NSB + 1;
 oggetto->tri[ia+_NSB].v[1].iv = ia + 1;
}




 cuci_vertici_iv( oggetto );



oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}

calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );

return( oggetto );
}





Oggetto3d* crea_cilindro( FTYPE h, FTYPE r, int _NSC)
{
int ia;
FTYPE a, sa;
FTYPE x, z;
FTYPE h2 = h / 2.0;
int iv, ip;
Oggetto3d* oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{
 oggetto = alloca_struct_oggetto();

 oggetto->num_vertici = _NSC*2 + 2;

 oggetto->num_poly = _NSC * 4;


 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);
}


sa = M_PI*2.0 / (FTYPE)_NSC;
a = 0.0;

set_localvert_oggetto( oggetto, 0,  0.0, h2, 0.0 );
iv = 1;


for(ia=0; ia<_NSC; ia++, a+=sa)
{
 x = cos( a ) * r;
 z = sin( a ) * r;
 set_localvert_oggetto( oggetto, iv,  x,  h2, z );
 set_localvert_oggetto( oggetto, iv + _NSC,  x, -h2, z );
 iv++;
}
iv = _NSC*2 + 1;
set_localvert_oggetto( oggetto, iv,  0.0, -h2, 0.0 );
iv++;



if(not_present)
{
ip = 0;

// chiusura superiore
for(ia=0; ia<_NSC; ia++)
{
 oggetto->tri[ip].v[0].iv = 0;
 oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC + 1;
 oggetto->tri[ip].v[2].iv = ia + 1;
 ip++;
}

// chiusura laterale
for(ia=0; ia<_NSC; ia++)
{
 oggetto->tri[ip].v[0].iv = ia + 1;
 oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC + 1;
 oggetto->tri[ip].v[2].iv = ia + _NSC + 1;
 ip++;
 oggetto->tri[ip].v[0].iv = (ia + 1)%_NSC + _NSC + 1;
 oggetto->tri[ip].v[1].iv = ia + _NSC + 1;
 oggetto->tri[ip].v[2].iv = (ia + 1)%_NSC + 1;
 ip++;
}

// chiusura inferiore
for(ia=0; ia<_NSC; ia++)
{
 oggetto->tri[ip].v[0].iv = _NSC*2 + 1;
 oggetto->tri[ip].v[1].iv = ia + _NSC + 1;
 oggetto->tri[ip].v[2].iv = (ia + 1)%_NSC + _NSC + 1;
 ip++;
}


 cuci_vertici_iv( oggetto );


oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}

calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );


return( oggetto );
}





Oggetto3d* crea_toroide( FTYPE r1, FTYPE r2, int _NSR, int _NSC, FTYPE n1, FTYPE n2)
{
FTYPE a1, sa1, a2, sa2;
FTYPE x, y, z, xt, yt, zt;
FTYPE sn, cs;
int iv, ip;
int ir, ia;
int ia1, ia2;
Oggetto3d *oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{

oggetto = alloca_struct_oggetto();


oggetto->num_vertici = _NSC * _NSR;

oggetto->num_poly = _NSR * _NSC*2;


alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);

}


a1=0.0;
sa1 = M_PI*2.0/(FTYPE)_NSR;

sa2 = M_PI*2.0/(FTYPE)_NSC;

iv = 0;
 for(ia1=0; ia1<_NSR; ia1++, a1+=sa1)
 {

  sn = pow_ok(sin( a1 ), n1);
  cs = pow_ok(cos( a1 ), n1);

  for(ia2=0, a2=0.0; ia2<_NSC; ia2++, a2+=sa2)
  {

    x = pow_ok(cos( a2 ), n2) * r2 - r1;
    y = pow_ok(sin( a2 ), n2) * r2;


   //z = 0.0;

   // rotazione sull'asse Y di 'a1' radianti
   xt = x * cs;//  -  z * sn;
   yt = y;
   zt = x * sn;//  +  z * cs;

    set_localvert_oggetto( oggetto, iv,  xt, yt, zt );
    iv++;
  }

}



if(not_present)
{
// calcola gli indici per i vertici dei poligoni
ip=0;

for( ir = 0; ir < _NSR; ir++ )
 for( ia = 0; ia < _NSC; ia++ )
 {
  oggetto->tri[ip].v[0].iv = (ia + 1)%_NSC  +  ir*_NSC;
  oggetto->tri[ip].v[1].iv = ia            +  ir*_NSC;
  oggetto->tri[ip].v[2].iv = (ia + 1)%_NSC  +  ((ir + 1)%_NSR) * _NSC;
  ip++;
  oggetto->tri[ip].v[0].iv = ia            +  ((ir + 1)%_NSR) * _NSC;
  oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC  +  ((ir + 1)%_NSR) * _NSC;
  oggetto->tri[ip].v[2].iv = ia            +  ir*_NSC;
  ip++;
 }



 cuci_vertici_iv( oggetto );


oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}

calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );


return( oggetto );
}





Oggetto3d* crea_sfera( FTYPE r, int _NSH, int _NSC)
{
FTYPE ang, sang, a, sa, h;
FTYPE r2 = r*r, rc;
FTYPE x, y, z;
int iv, ip;
int ih, ia, iang;
int tot_poly = (_NSC*2)  +  (_NSH-2) * (_NSC*2);
Oggetto3d* oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{
oggetto = alloca_struct_oggetto();

oggetto->num_vertici = _NSC*(_NSH-2) + 2;

oggetto->num_poly = tot_poly;


 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);

}


//  numero poly di chiusura    numero poly di ogni fascia * numero fasce
//               |                  |
//               V                  V
// tot_poly  =  (_NSC*2)  +  (_NSH-2) * (_NSC*2)
//


sang = M_PI / (FTYPE)(_NSH-1);
sa = M_PI*2.0 / (FTYPE)_NSC;

set_localvert_oggetto( oggetto, 0, 0.0, r, 0.0 );
iv = 1;

 for(iang=1, ang=0+sang; iang<_NSH-1; iang++, ang+=sang)
 {
  h = cos( ang ) * r;
  rc = sin( ang ) * r;

  for(ia=0, a=0.0; ia<_NSC; ia++, a+=sa)
  {
   x = cos( a ) * rc;
   y = h;
   z = sin( a ) * rc;
   set_localvert_oggetto( oggetto, iv, x, y, z );
   iv++;
  }
 }

set_localvert_oggetto( oggetto, iv, 0.0, -r, 0.0 );
iv++;


if(not_present)
{

 //oggetto->num_vertici = iv;


// ecco come sono numerati i vertici delle chiusure superiori e inferiori:
//       0
//	    /\
//     /  \
//   1/____\2
//
//   2 ____ 1
//    \    /
//     \  /
//	    \/
//	     0


ip = 0;

// chiusura superiore
for(ia=0; ia<_NSC; ia++)
{
 oggetto->tri[ip].v[0].iv = 0;
 oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC + 1;
 oggetto->tri[ip].v[2].iv = ia + 1;
 ip++;
}


// ecco come sono numerati i vertici laterali:
//
//   0 ___2
//    |  /|1
//    | / |
//   1|/__|
//     2   0

if(_NSH > 3)
for(ih=0; ih<_NSH-3; ih++)
 for(ia=0; ia<_NSC; ia++)
 {
  oggetto->tri[ip].v[0].iv = ia  +  (ih)*_NSC + 1;
  oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC  +  (ih)*_NSC + 1;
  oggetto->tri[ip].v[2].iv = ia  +  (ih + 1)*_NSC + 1;
  ip++;

  oggetto->tri[ip].v[0].iv = (ia + 1)%_NSC  +  (ih + 1)*_NSC + 1;
  oggetto->tri[ip].v[1].iv = ia  +  (ih + 1)*_NSC + 1;
  oggetto->tri[ip].v[2].iv = (ia + 1)%_NSC  +  (ih)*_NSC + 1;
  ip++;
 }


// chiusura inferiore
iv = oggetto->num_vertici-1;

for(ia=0; ia<_NSC; ia++)
{
 oggetto->tri[ip].v[0].iv = iv;
 oggetto->tri[ip].v[1].iv = iv - _NSC + ia;
 oggetto->tri[ip].v[2].iv = iv - _NSC + (ia + 1)%_NSC;
 ip++;
}


 cuci_vertici_iv( oggetto );

oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}


calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );


oggetto->raggio_sfera = r;//calcola_sfera_oggetto( oggetto );


return( oggetto);
}






Oggetto3d* crea_tubo( FTYPE P, FTYPE Q, FTYPE R,
                      FTYPE r1, FTYPE r2, int _NSL, int _NSC, int nuovo)
{
int il, ia, ip, ir;
FTYPE al, sl, a, sa;
FTYPE px, py, pz;
Punto3d p1, p2, p3;
FTYPE x, y, z, xt, yt, zt;
FTYPE ax, ay, az;
int iv;
Vettore3d u, v, w, vb;
Oggetto3d* oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{
 oggetto = alloca_struct_oggetto();

 oggetto->num_vertici = _NSC * _NSL;

 oggetto->num_poly = _NSL * _NSC*2;

 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);
}


sl = M_PI*2.0 / (FTYPE)_NSL;

sa = M_PI*2.0 / (FTYPE)_NSC;
iv=0;

for(il=0, al=0.0; il<_NSL; il++, al-=sl)
{
 p1.x = cos((al-sl) * P) * r1;
 p1.y = sin((al-sl) * Q) * r1;
 p1.z = sin((al-sl) * R) * r1;

 p2.x = cos(al * P) * r1;
 p2.y = sin(al * Q) * r1;
 p2.z = sin(al * R) * r1;

 p3.x = cos((al+sl) * P) * r1;
 p3.y = sin((al+sl) * Q) * r1;
 p3.z = sin((al+sl) * R) * r1;



// Calcola i vettori per la nuova base

 crea_vettore(&p1, &p2,  &w);
 crea_vettore(&p1, &p3,  &vb);
 prodotto_vettoriale(&w, &vb,  &u);
 prodotto_vettoriale(&w, &u,   &v);
 normalizza_vettore( &u );
 normalizza_vettore( &v );

 for(ia=0, a=0.0; ia<_NSC; ia++, a+=sa)
 {
  // Calcola le coordinate dei punti nella base {(1,0,0);(0,1,0);(0,0,1)}
  x = cos(a) * r2;
  y = sin(a) * r2*0.7;
  z = 0.0;

// Cambiamento di base per ogni punto

  xt = x*u.x + y*v.x + z*w.x  + p2.x;
  yt = x*u.y + y*v.y + z*w.y  + p2.y;
  zt = x*u.z + y*v.z + z*w.z  + p2.z;

  set_localvert_oggetto(oggetto, iv,  xt, yt, zt);
  iv++;

 } // end "ia"

} // end  "il"




if(not_present)
{

// calcola gli indici per i vertici dei poligoni
ip=0;

for( ir = 0; ir < _NSL; ir++ )
 for( ia = 0; ia < _NSC; ia++ )
 {
  oggetto->tri[ip].v[0].iv = (ia + 1)%_NSC  +  ir*_NSC;
  oggetto->tri[ip].v[1].iv = ia            +  ir*_NSC;
  oggetto->tri[ip].v[2].iv = (ia + 1)%_NSC  +  ((ir + 1)%_NSL) * _NSC;
  ip++;
  oggetto->tri[ip].v[0].iv = ia            +  ((ir + 1)%_NSL) * _NSC;
  oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC  +  ((ir + 1)%_NSL) * _NSC;
  oggetto->tri[ip].v[2].iv = ia            +  ir*_NSC;
  ip++;
 }


 cuci_vertici_iv( oggetto );

oggetto->aperto = FALSO;

oggetto->interno = FALSO;

}


calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );


return( oggetto );
}







Oggetto3d* crea_tubo_pts( Punto3d *pts, int np, Punto3d *shape, int _NSC, FTYPE r2, int nl)
{
int i, ia, ip, ir;
FTYPE a, sa;
FTYPE px, py, pz;
Punto3d p1, p2, p3;
FTYPE x, y, z, xt, yt, zt;
int iv;
Vettore3d u, v, w, vb;

BezCurve bez_curve;
FTYPE t, dt, dt1, dt2;

Oggetto3d* oggetto;
int not_present = 0;



if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{
 oggetto = alloca_struct_oggetto();

 oggetto->num_vertici = _NSC * nl;

 oggetto->num_poly = nl * _NSC*2;

 alloca_memoria_oggetto( oggetto, oggetto->num_vertici, oggetto->num_poly );
}

iv = 0;

sa = M_PI*2.0 / (FTYPE)_NSC;


Bez_create_curve(&bez_curve, pts, np, TRUE);
t = 0.0;
dt = 100.0 / (FTYPE)nl;
dt2 = dt  / 2.0;
dt1 = dt2 / 2.0;


for(i=0; i<nl; i++, t+=dt)
{
Bez_get_point_coord_percent(&bez_curve, t,       &p1);
Bez_get_point_coord_percent(&bez_curve, t + dt1, &p2);
Bez_get_point_coord_percent(&bez_curve, t + dt2, &p3);
/*
 p1.x = pts[i].x;
 p1.y = pts[i].y;
 p1.z = pts[i].z;

 p2.x = pts[(i+1)%np].x;
 p2.y = pts[(i+1)%np].y;
 p2.z = pts[(i+1)%np].z;

 p3.x = pts[(i+2)%np].x;
 p3.y = pts[(i+2)%np].y;
 p3.z = pts[(i+2)%np].z;
*/

// Calcola i vettori per la nuova base
 crea_vettore(&p1, &p2,  &w);
 crea_vettore(&p1, &p3,  &vb);
 prodotto_vettoriale(&w, &vb,  &u);
 prodotto_vettoriale(&w, &u,   &v);
  normalizza_vettore( &u );
  normalizza_vettore( &v );
  normalizza_vettore( &w );

 for(ia=0, a=0.0; ia<_NSC; ia++, a+=sa)
 {
  // Calcola le coordinate dei punti nella base {(1,0,0);(0,1,0);(0,0,1)}
  if(shape==NULL)
  {
   x = cos(a) * r2;
   y = sin(a) * r2;
   z = 0.0;
  }
   else
  {
   x = shape[ia].x * r2;
   y = shape[ia].y * r2;
   z = shape[ia].z * r2;
  }
// Cambiamento di base per ogni punto

  xt = x*u.x + y*v.x /*+ z*w.x*/  + p1.x;
  yt = x*u.y + y*v.y /*+ z*w.y*/  + p1.y;
  zt = x*u.z + y*v.z /*+ z*w.z*/  + p1.z;

  set_localvert_oggetto(oggetto, iv,  xt, yt, zt);
  iv++;

 } // end "ia"

} // end  "i"




if(not_present)
{

// calcola gli indici per i vertici dei poligoni
ip=0;

for( ir = 0; ir < nl; ir++ )
 for( ia = 0; ia < _NSC; ia++ )
 {
  oggetto->tri[ip].v[0].iv = (ia + 1)%_NSC  +  ir*_NSC;
  oggetto->tri[ip].v[2].iv = ia            +  ir*_NSC;
  oggetto->tri[ip].v[1].iv = (ia + 1)%_NSC  +  ((ir + 1)%nl) * _NSC;
  ip++;
  oggetto->tri[ip].v[0].iv = ia            +  ((ir + 1)%nl) * _NSC;
  oggetto->tri[ip].v[2].iv = (ia + 1)%_NSC  +  ((ir + 1)%nl) * _NSC;
  oggetto->tri[ip].v[1].iv = ia            +  ir*_NSC;
  ip++;
 }


 cuci_vertici_iv( oggetto );

oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}

calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );

return( oggetto );
}




Oggetto3d* crea_tornito( FTYPE *pts, int NP, int _NSC)
{
FTYPE x, y, z, xt, yt, zt;
FTYPE sn, cs;
int iv, ip;
int ir, ia;
FTYPE a, sa;
Oggetto3d* oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{
 oggetto = alloca_struct_oggetto();

 oggetto->num_vertici = _NSC * NP;

 oggetto->num_poly = NP * _NSC*2;

 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);
}

iv = 0;

a=0.0;
sa = M_PI*2.0/(FTYPE)_NSC;

 for(ia=0; ia<_NSC; ia++, a+=sa)
 {
  sn = sin( a );
  cs = cos( a );

  for(ip=0; ip<NP*3; ip+=3)
  {
   x = pts[ip];
   y = pts[ip+1];
   z = pts[ip+2];
   // rotazione sull'asse Y di 'a' radianti
   xt = x * cs  -  z * sn;
   yt = y;
   zt = x * sn  +  z * cs;

    set_localvert_oggetto( oggetto, iv,  xt, yt, zt );
    iv++;
  }

}


if(not_present)
{
// calcola gli indici per i vertici dei poligoni
ip=0;

for( ia = 0; ia < _NSC; ia++ )
 for( iv = 0; iv < NP-1; iv++ )
 {
  oggetto->tri[ip].v[0].iv = ia*NP + iv;
  oggetto->tri[ip].v[1].iv = ia*NP + iv+1;
  oggetto->tri[ip].v[2].iv = ((ia+1)%_NSC) * NP + iv;
  ip++;
  oggetto->tri[ip].v[0].iv = ((ia+1)%_NSC) * NP  + iv+1;
  oggetto->tri[ip].v[1].iv = ((ia+1)%_NSC) * NP  + iv;
  oggetto->tri[ip].v[2].iv = ia*NP + iv+1;
  ip++;
 }


 cuci_vertici_iv( oggetto );

oggetto->aperto = FALSO;

oggetto->interno = FALSO;
}

calcola_normali_oggetto( oggetto );

calcola_coordinate_mapping_oggetto( oggetto );

oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );

return( oggetto );
}









// crea_griglia()->parametri:
// l, h = larghezza, altezza
// _NSL, _NSH = numero di suddivisioni lungo l e h
// wx, wy = fattore di scala della texture lungo l e h
Oggetto3d* crea_griglia( FTYPE l, FTYPE h, int _NSL, int _NSH, FTYPE wx, FTYPE wy)
{
FTYPE x, y;
FTYPE dx, dy;
int i, j;
int iv, ip;
int _NVL, _NVH;
FTYPE u, v, du, dv;
Oggetto3d* oggetto;
int not_present = 0;


if( (oggetto = get_current_object())==NULL ) not_present = 1;

if( not_present )
{

oggetto = alloca_struct_oggetto();

_NVL = _NSL + 1;
_NVH = _NSH + 1;

oggetto->num_vertici = _NVL * _NVH;

oggetto->num_poly = (_NSL * _NSH) * 2;


 alloca_memoria_oggetto(oggetto, oggetto->num_vertici, oggetto->num_poly);
}

du = 1.0 / (FTYPE)_NSL;
dv = 1.0 / (FTYPE)_NSH;

dx = l * du;
dy = h * dv;

iv=0;
ip=0;


du*=wx;
dv*=wy;


for(j=0, v=0, y=-h/2.0; j<_NVH; j++)
{

for(i=0, u=0, x=-l/2.0; i<_NVL; i++)
 {
    set_localvert_oggetto( oggetto, iv,  x, 0, y );

    set_normale_oggetto( oggetto, iv,  0, 1, 0 );

    set_texcoord_oggetto( oggetto, iv,  u, v );

/*
1---2
|\  |
|  \|
3---4
*/
if(not_present)
  if( i!=_NSL && j!=_NSH )
  {


    /* 0-2
       1/. */
     oggetto->tri[ip].v[0].iv = iv;
     oggetto->tri[ip].v[1].iv = iv + 1;
     oggetto->tri[ip].v[2].iv = iv + _NVL;

     ip++;
    /* ./0
       1-2  */
     oggetto->tri[ip].v[0].iv = iv + 1;
     oggetto->tri[ip].v[1].iv = iv + 1 + _NVL;
     oggetto->tri[ip].v[2].iv = iv + _NVL;

    ip++;

  }


    x+=dx;
    u+=du;
    iv++;
 }

    y+=dy;
    v+=dv;

}

if(not_present) cuci_vertici_iv( oggetto );

/*************/

 oggetto->raggio_sfera = calcola_sfera_oggetto( oggetto );

 oggetto->aperto = VERO;

 oggetto->interno = FALSO;

return( oggetto );
}



//////////////////////////////////////
//////////////////////////////////////
//////////////////////////////////////


void rumore_oggetto( Oggetto3d *src, Oggetto3d *dest, FTYPE ix, FTYPE iy, FTYPE iz, int seed )
{
int i;

if(!(dest && src)) 
{
 MessageBox(NULL, "Oggetti non allocati: skippo", "rumore_oggetto()", MB_OK);
 return;
}

srand(seed);

 if(ix) for(i=0; i<src->num_vertici; i++) 
   dest->local_vert[i].x = src->local_vert[i].x + ((rand()%10001-5000)/5000.0) * ix;

 if(iy) for(i=0; i<src->num_vertici; i++) 
   dest->local_vert[i].y = src->local_vert[i].y + ((rand()%10001-5000)/5000.0) * iy;

 if(iz) for(i=0; i<src->num_vertici; i++) 
   dest->local_vert[i].z = src->local_vert[i].z + ((rand()%10001-5000)/5000.0) * iz;

 if(ix || iy || iz)  dest->raggio_sfera = calcola_sfera_oggetto( dest );

 calcola_normali_oggetto( dest );
}




void rumore_radiale_oggetto( Oggetto3d *src, Oggetto3d *dest, FTYPE pwr, FTYPE cx, FTYPE cy, FTYPE cz, int seed )
{
int i;
Vettore3d ris;
Punto3d c;

if( dest == NULL || src == NULL || pwr==0.0 ) 
{
// MessageBox(NULL, "rumore_radiale_oggetto()", "Oggetti non allocati: skippo", MB_OK);
 return;
}

c.x = cx;
c.y = cy;
c.z = cz;

srand(seed);

for(i=0; i < src->num_vertici; i++) 
{    
 crea_vettore( &c, &src->local_vert[i], &ris );
 normalizza_vettore_k( &ris, ((rand()%10001-5000) / 5000.0) * pwr );

 dest->local_vert[i].x = src->local_vert[i].x  +  ris.x;
 dest->local_vert[i].y = src->local_vert[i].y  +  ris.y;
 dest->local_vert[i].z = src->local_vert[i].z  +  ris.z;
}

    
  calcola_normali_oggetto( dest );
  //if(M3D_CalculateSphere) dest->raggio_sfera = calcola_sfera_oggetto( dest );
 
if(src->next_ogg != NULL && dest->next_ogg != NULL)

rumore_radiale_oggetto( (Oggetto3d *)src->next_ogg, 
                        (Oggetto3d *)dest->next_ogg, pwr, c.x, c.y, c.z, seed );


}


void normalizza_oggetto(Oggetto3d *oggetto, FTYPE dim )
{
FTYPE s, raggio_sfera;
Oggetto3d *cur_ogg;

if(!oggetto) 
{
// MessageBox(NULL, "normalizza_oggetto()", "Oggetti non allocati: skippo", MB_OK);
 return;
}

// trova il massimo dei raggi:

 cur_ogg = oggetto;

 raggio_sfera = cur_ogg->raggio_sfera;

 while( cur_ogg->next_ogg != NULL )
 {
  cur_ogg = (Oggetto3d *)cur_ogg->next_ogg;

  if( cur_ogg->raggio_sfera > raggio_sfera )  raggio_sfera = cur_ogg->raggio_sfera;
 }

  if(raggio_sfera) 
  {
   s = dim / raggio_sfera;
   scala_oggetto(oggetto, s, s, s);
  }

}



#endif