/*

  Ray 1.0 - Recursive Noisy 3D textures! (i.e.: Ken Perlin noise)

*/
 



#ifndef _LIB_NOISE
#define _LIB_NOISE




// the 3D texturing-cube dimensions
#define DIM 2.0

// number of cells of the 3D cube (NUM x NUM x NUM) forming the LF noisy texture
// to avoid artifacts should be 'floor(RAND_DIM ^ 1/3)'
#define NUM 15

// dimension of the lookup table
#define RAND_DIM (1<<12)
//#define RAND_DIM  ((NUM)*(NUM)*(NUM))




// random values look-up table
FTYPE g_rand_vals[ RAND_DIM  +  NUM*NUM+1 ]; 

#define G_RAND(i) g_rand_vals[ (int)(i)&(RAND_DIM-1) ]

#define G_RAND3(i,j,k) g_rand_vals[ ((i)+NUM*((j)+(k)*NUM)) % (RAND_DIM) ]


void init_rand()
{
int i, j, k;

 //for(i=0; i<RAND_DIM; i++) g_rand_vals[i] = ((rand()&0xFFFFFF) - 0x7FFFFF)/(FTYPE)0x7FFFFF;
 for(i=0; i<RAND_DIM; i++) g_rand_vals[i] = (rand()&0xFFFFFF)/(FTYPE)0xFFFFFF;

 // 3D wrapping lookup-tweak! ;)

 for(i=0; i<NUM; i++)
  for(j=0; j<NUM; j++)
  {
   G_RAND3(i,j,NUM-1)=G_RAND3(i,j,0);
	 G_RAND3(NUM-1,i,j)=G_RAND3(0,i,j);
	 G_RAND3(i,NUM-1,j)=G_RAND3(i,0,j);	 
  }



}


FTYPE cubic(double u)
{
#define _Y_SCALE  (M_SQRT1_3 * 2.0)
#define _X_SCALE  (M_SQRT1_3 * 2.0)

u = (u - .5) * _X_SCALE;
return
 (u*(1.0-u*u)) * _Y_SCALE + 0.5;
}

// returns a [-1.0 - 1.0] normalized value
FTYPE my_noise3d(FTYPE x, FTYPE y, FTYPE z, int n)
{

FTYPE u, v, w; // position in the cell
int   i, j, k; // index of the cell
FTYPE a, b, t1, t2;
int off;
/*
x+=DIM*2 + .1f;
y+=DIM*2 + .1f;
z+=DIM*2 + .1f;
*/

if(x < 0) x=-x;
if(y < 0) y=-y;
if(z < 0) z=-z;


// select the needed cell to interpolate:
i = fmodf(x, DIM) * ((NUM-1) / DIM);
j = fmodf(y, DIM) * ((NUM-1) / DIM);
k = fmodf(z, DIM) * ((NUM-1) / DIM);


off = i + NUM*(j + k*NUM);


// select interpolation coefficients along the 3 cell axis:
u = fmodf(x * ((NUM-1)/DIM), 1.0);
v = fmodf(y * ((NUM-1)/DIM), 1.0);
w = fmodf(z * ((NUM-1)/DIM), 1.0);


// cubic transformation: half-way between linear and cosine
u = cubic(u);
v = cubic(v);
w = cubic(w);


/*
// cosine interpolation: looks nicer
u = .5f - cosf(u * M_PI)*.5f;
v = .5f - cosf(v * M_PI)*.5f;
w = .5f - cosf(w * M_PI)*.5f;
*/


// Trilinear interpolation of cell-values:

// Front x-y plane
a  = G_RAND(off + 0) + (G_RAND(off + 1) - G_RAND(off + 0)) * u;
off+=NUM;
b  = G_RAND(off + 0) + (G_RAND(off + 1) - G_RAND(off + 0)) * u;
t1 = a + (b - a) * v;

off += NUM*NUM - NUM;
// Back x-y plane
a  = G_RAND(off + 0) + (G_RAND(off + 1) - G_RAND(off + 0)) * u;
off+=NUM;
b  = G_RAND(off + 0) + (G_RAND(off + 1) - G_RAND(off + 0)) * u;
t2 = a + (b - a) * v;

// final interpolation:
if(n<=0)
 return  (t1 + (t2 - t1) * w);

return  
((t1 + (t2 - t1) * w)*.75 +
 // transforming arguments for each recursion gives nicer results
 my_noise3d(y*2.0, -z*2.0, x*2.0, n-1)*.25 );
  //my_noise3d((y+z)*2.0f, z, (x-z)*2.0f, n-1)*.25 );
  //my_noise3d(y*2.0f, z*2.0f, x*2.0f, n-1)*.25 );
  //my_noise3d(-y+x*2.0f, x+y*2.0f, z*2.0f, n-1)*.25 );

}



#endif

