/* _Copyright (c) 1988 Regents of the University of _California */

#ifndef _LIB_PERLIN
#define _LIB_PERLIN

/*
 *  perlin.h - noise functions for random textures.
 *
 *     _Credit for the smooth algorithm goes to Ken Perlin.
 *     (ref. SIGGR_APH Vol 19, No 3, pp 287-96)
 *
 *     4/15/86
 *     5/19/88	_Added fractal noise function
 */

#include <math.h>
#include <stdlib.h>

#define  _A		0
#define  _B		1
#define  _C		2
#define  _D		3

#define  RAND3a(x,y,z)	frand(67*(x)+59*(y)+71*(z))
#define  RAND3b(x,y,z)	frand(73*(x)+79*(y)+83*(z))
#define  RAND3c(x,y,z)	frand(89*(x)+97*(y)+101*(z))
#define  RAND3d(x,y,z)	frand(103*(x)+107*(y)+109*(z))

/* hermite interpolant polynomyals: */
#define  HPOLY1(t)	((2.0*t-3.0)*t*t+1.0)
#define  HPOLY2(t)	((-2.0*t+3.0)*t*t)
#define  HPOLY3(t)	(((t-2.0)*t+1.0)*t)
#define  HPOLY4(t)	((t-1.0)*t*t)

//double  *noise3(), fnoise3(), frand();

//static  interpolate();

static long   g_xlim[3][2];
static double g_xarg[3];

#define  EPSILON	0.01		/* error allowed in fractal */

#define  fRAND3(x,y,z)	frand(17*(x)+23*(y)+29*(z))



double frand(register long s)			/* get random number from seed */
{
	s = s<<13 ^ s;
	return(1.0-((s*(s*s*15731+789221)+1376312589)&0x7fffffff)/1073741824.0);
}

void interpolate(double *f, register int i, register int n)
{
double  f0[4], f1[4], hp1, hp2;
double t1,t2,t3;



	if (n == 0) 
	{
		t1 = g_xlim[0][i&1];
		t2 = g_xlim[1][i>>1&1];
		t3 = g_xlim[2][i>>2];

		f[_A] = RAND3a(t1, t2, t3);
		f[_B] = RAND3b(t1, t2, t3);
		f[_C] = RAND3c(t1, t2, t3);
		f[_D] = RAND3d(t1, t2, t3);
	} 
	  else 
	{
		n--;
		interpolate(f0, i, n);
		interpolate(f1, i | 1<<n, n);

		hp1 = HPOLY1(g_xarg[n]); 
		hp2 = HPOLY2(g_xarg[n]);
		f[_A] = f0[_A]*hp1 + f1[_A]*hp2;
		f[_B] = f0[_B]*hp1 + f1[_B]*hp2;
		f[_C] = f0[_C]*hp1 + f1[_C]*hp2;
		f[_D] = f0[_D]*hp1 + f1[_D]*hp2 +
		       f0[n]*HPOLY3(g_xarg[n]) +
			   f1[n]*HPOLY4(g_xarg[n]);
	}
}


double *noise3( register double  *xnew )			/* compute the noise function */

{
	static double  x[3] = {-100000.0, -100000.0, -100000.0};
	static double  f[4];

	if (x[0]==xnew[0] && x[1]==xnew[1] && x[2]==xnew[2])
		return(f);

	x[0] = xnew[0]; 
	x[1] = xnew[1]; 
	x[2] = xnew[2];

	g_xlim[0][0] = floor(x[0]); 
	g_xlim[0][1] = g_xlim[0][0] + 1;

	g_xlim[1][0] = floor(x[1]); 
	g_xlim[1][1] = g_xlim[1][0] + 1;

	g_xlim[2][0] = floor(x[2]); 
	g_xlim[2][1] = g_xlim[2][0] + 1;

	g_xarg[0] = x[0] - g_xlim[0][0];
	g_xarg[1] = x[1] - g_xlim[1][0];
	g_xarg[2] = x[2] - g_xlim[2][0];

	interpolate(f, 0, 3);

	return(f);
}




double perlin3(double *p)			/* compute fractal noise function */
{
	long  t[3], v[3], beg[3];
	double  fval[8], fc;
	int  branch;
	register long  s;
	register int  i, j;

	/* get starting cube */
	s = (long)(1.0 / EPSILON);

	for (i = 0; i < 3; i++)
	{
		t[i] = s*p[i];
		beg[i] = s*floor(p[i]);
	}

	for (j = 0; j < 8; j++) 
	{
		for (i = 0; i < 3; i++) 
		{
			v[i] = beg[i];
			if (j & 1<<i) v[i] += s;
		}

		fval[j] = fRAND3(v[0],v[1],v[2]);
	}

	/* compute fractal */
	for (;;) 
	{
		fc = 0.0;

		for (j = 0; j < 8; j++) fc += fval[j];

		fc *= 0.125;

		if ((s >>= 1) == 0)	return(fc);		/* close enough */

		branch = 0;

		for (i = 0; i < 3; i++) 
		{	/* do center */
			v[i] = beg[i] + s;

			if (t[i] > v[i]) 
			{
				branch |= 1<<i;
			}
		}

		fc += s * EPSILON * fRAND3(v[0],v[1],v[2]);
		fval[~branch & 7] = fc;

		for (i = 0; i < 3; i++) 
		{	/* do faces */
			if (branch & 1<<i)
				v[i] += s;
			else
				v[i] -= s;

			fc = 0.0;
			for (j = 0; j < 8; j++)
				if (~(j^branch) & 1<<i)	fc += fval[j];

			fc = 0.25*fc + s * EPSILON * fRAND3(v[0],v[1],v[2]);
			fval[~(branch^1<<i) & 7] = fc;

			v[i] = beg[i] + s;
		}

		for (i = 0; i < 3; i++) 
		{	/* do edges */
			j = (i+1)%3;
			if (branch & 1<<j)
				v[j] += s;
			else
				v[j] -= s;
			j = (i+2)%3;
			if (branch & 1<<j)
				v[j] += s;
			else
				v[j] -= s;

			fc = fval[branch & ~(1<<i)];
			fc += fval[branch | 1<<i];
			fc = 0.5*fc + s * EPSILON * fRAND3(v[0],v[1],v[2]);

			fval[branch^1<<i] = fc;

			j = (i+1)%3;
			v[j] = beg[j] + s;

			j = (i+2)%3;
			v[j] = beg[j] + s;
		}

		for (i = 0; i < 3; i++)		/* new cube */
			if (branch & 1<<i) beg[i] += s;

	}
}

#endif