/*

 Ray 1.0 - Ray Tracing computational core

*/

#ifndef _LIB_RAYTRACER
#define _LIB_RAYTRACER


#include "mathconsts.h"
#include "enum.h"
#include "log.h"
#include "math3d.h"

#include "perlin.h"
#include "noise.h"

#include "solvers.h"

#include "colors.h"
#include "lights.h"
#include "materials.h"

#include "camera.h"
#include "scene.h"

#include "draw24.h"
#include "SDLwindow.h"



int g_i,g_j;
bool g_clip = false;

t_intersection *find_min_t(t_ray *ray0, t_scene *sc, int id_sh, FTYPE t_max)
{
int i;
static int unknown;
static t_shape *sh;
t_ray *ray;

/******/
 reset_intersections( sc->clip_far );
/******/

for(i=0; i < sc->num_objects; i++)
{

sh = sc->shapes[i];

if( g_clip && !sh->clipping_rect.covers_all )
{
 if( sh->clipping_rect.visible == false ) continue;

 if( OUTSIDE( g_i,  sh->clipping_rect.x_min, sh->clipping_rect.x_max )) continue;
 if( OUTSIDE( g_j,  sh->clipping_rect.y_min, sh->clipping_rect.y_max )) continue;
}



//
// Apply shape's inverse transformation matrix to incoming ray
//
ray = &sh->ray_inv;

m_transform_ray( sh->m_inv, ray0, ray );

//
// Intersect the transformed ray with shape:
//
 switch( sh->type )
 {
 /* i.e.:"  case X: solve_x( ray, sh );  break;   "*/

// --- Planar shapes: ---
  case PLANE:    solve_plane( ray, sh );  break;

  case TRIANGLE: solve_triangle( ray, sh );  break;

  case BOX:      solve_box( ray, sh );  break;


// --- Quadric shapes: ---
/*
  case SPHERE: case CYLINDER: case CONE:
*/
  case QUADRIC: solve_quadric( ray, sh );  break;


// --- Cubic shapes: ---
	case TRILINEAR_CELL:  solve_trilinear_cell( ray, sh ); break;
  case CUBIC: break;

// --- Quartic shapes: ---
	case BOX4:     solve_box4( ray, sh );  break;
  case TORUS:    solve_torus( ray, sh, SOLVE_ALL );  break;
	case TANGLE:   solve_tangle( ray, sh );  break;
  case QUARTIC:  solve_quartic( ray, sh ); break;


  case METABALLS:
  case CSG:  break;


 default: 
  g_unknown++;
 break;
 }
/*TODO*/
if(id_sh) 
 if(g_t_min_ptr!=NULL)
	if(g_t1 < t_max) break;

}

return g_t_min_ptr;

}






t_intersection *cast_shadows( t_scene *sc, t_vector3d *to_l, FTYPE t_max, FTYPE ox, FTYPE oy, FTYPE oz)
{
t_ray test;

test.d = *to_l;

// needed only if distance is necessary
//  norm_v(&test.d);

test.o.x = ox + to_l->x * EPSILON_INT;
test.o.y = oy + to_l->y * EPSILON_INT;
test.o.z = oz + to_l->z * EPSILON_INT;

 find_min_t( &test, sc, 1, t_max );

 return g_t_min_ptr;
}









/* 1.0 is totally exposed, 0.0 is totally occluded */
FTYPE cast_soft_shadows( t_scene *sc, t_vector3d *to_l, FTYPE ox, FTYPE oy, FTYPE oz, t_light *l)
{
t_ray test;
int i, j;
t_vector3d du, dv;
t_point3d pv;
FTYPE sy, sx;
int hits;

t_billboard_map *map = &l->billboard_map;


// Orient the billboarding map for the specific 'to light' vector:
orient_billboard_map( map, to_l ); // (no matter if 'to_l' isn't reversed)


test.o.x = ox + to_l->x;// * EPSILON_INT;
test.o.y = oy + to_l->y;// * EPSILON_INT;
test.o.z = oz + to_l->z;// * EPSILON_INT;

sy = map->start_y;

du.x = map->b.u.x * map->delta;
du.y = map->b.u.y * map->delta;
du.z = map->b.u.z * map->delta;

dv.x = map->b.v.x * map->delta;
dv.y = map->b.v.y * map->delta;
dv.z = map->b.v.z * map->delta;

pv.x = sy * map->b.v.x + l->pos.x - ox;
pv.y = sy * map->b.v.y + l->pos.y - oy;
pv.z = sy * map->b.v.z + l->pos.z - oz;

hits = 0;

for(j=0; j < map->height; j++)
{
 sx = map->start_x[j];

 test.d.x  =  sx * map->b.u.x + pv.x;
 test.d.y  =  sx * map->b.u.y + pv.y;
 test.d.z  =  sx * map->b.u.z + pv.z;

 add_p_v( &pv, &dv ); 
/*
 Unoptimized version:
 test.d.x  =  sx * map->b.u.x  +  sy * map->b.v.x + l->pos.x - ox;
 test.d.y  =  sx * map->b.u.y  +  sy * map->b.v.y + l->pos.y - oy;
 test.d.z  =  sx * map->b.u.z  +  sy * map->b.v.z + l->pos.z - oz;
 sy += map->delta;
*/
	for(i=0; i < map->steps[j]; i++)
	{
     //g_clip = false;

		find_min_t( &test, sc, 1, sc->clip_far );
		 
		 if( g_t_min_ptr == NULL )  hits++;

		test.d.x += du.x;
		test.d.y += du.y;
		test.d.z += du.z;
	}

}


return (FTYPE)hits * map->inv_area;

}






void get_texture_value( FTYPE *r, FTYPE *g, FTYPE *b, 
											  t_shape *sh, FTYPE x, FTYPE y, FTYPE z )
{
t_material *mat = sh->material;

 switch( mat->tex )
 {
 case TEX_NONE: break;

  case TEX_XOR:  *r = *g = *b = XOR3D(x,y,z); break;

  case TEX_PALETTE: color_rainbow(r, g, b, XOR3D(x,y,z));  break;

  case TEX_PERLIN:  color_rainbow(r, g, b, my_noise3d(x,y,z,0)); break;

  case TEX_CHECKER: *r = *g = *b = CHECKER3D(x,y,z); break;

  case TEX_IMAGE: break;

  default:
	 
   color_rainbow(r, g, b, XOR3D(x,y,z));

 break;
}

}




void get_texture_coord( FIX_TYPE *ut, FIX_TYPE *vt, t_shape *sh, 
											  FTYPE x, FTYPE y, FTYPE z)
{
FTYPE u, v;

 switch( sh->type )
 {
 case QUADRIC:
  u = atan2f(z,x)*(1.0/M_PI);
  v = atan2f(y,x)*(1.0/M_PI);
	break;
 case TRILINEAR_CELL:
	u = x;
  v = z;
	break;
 case BOX4:
  u = atan2f(z,x)*(1.0/M_PI);
  v = atan2f(y,x)*(1.0/M_PI);
	break;
 case TORUS:  /*TODO*/
	u = x;
    v = z;
	break;
 case TANGLE:  
	u = x*.25;
	v = y*.25;
	break;
 //case QUARTIC:  
	
   default:
		u = x;
		v = z;
  break;
 }

if(u < 0) u=-u;
if(v < 0) v=-v;

*ut = FLOAT_2_FIX256( u );
*vt = FLOAT_2_FIX256( v );

}





void get_shape_normal(t_shape *sh, t_vector3d *n, float x, float y, float z)
{
 switch( sh->type )
 {
 case QUADRIC:
       get_quadric_normal( n, sh, x,y,z );
			 break;


 case TRILINEAR_CELL:
	     get_trilinear_cell_normal( n, sh, x,y,z );
       break;


 case BOX4:
	     get_box4_normal( n, sh, x,y,z );
       break;

 case TORUS:  
	     get_torus_normal( n, sh, x,y,z );
       break;

 case TANGLE:  
	     get_tangle_normal( n, sh, x,y,z );
       break;

 case QUARTIC:  
	     get_quartic_normal( n, sh, x,y,z );
       break;


 default:
  *n = g_t_min_ptr->n;
 break;
}

}









bool get_lighting(t_rgb *diff, t_rgb *spec,
									t_scene *sc, t_shape *sh, t_ray *ray, 
									FTYPE x, FTYPE y, FTYPE z, t_vector3d *n, int recursion)
{
FTYPE lr, lg, lb;
FTYPE sr, sg, sb;
FTYPE dist2, dist, t, t1;

FTYPE cos_alpha; // [Diffuse ] (cosine of the) angle between incident light-ray and normal
FTYPE cos_beta;  // [Specular] (cosine of the) angle between point->observer and reflected light-ray
FTYPE specular;

FTYPE att; // to track the amount of transmitted light energy
bool inside = false; // flag for inside-object-watching test
t_vector3d to_l, ref;
t_material *mat = sh->material;
t_light *light;




// Now we have:
// - the (nearest) point's position: ('x, y, z')
// - it's normal           ('n.x, n.y, n.z')
// - the shape to wich the point belongs and it's properties ('g_t_min_ptr->shape_ptr')
// - it's distance from the viewpoint ('g_t_min')



////////////////////////////////
///* Do light calculations: *///
////////////////////////////////

// reset accumulators
lr = lg = lb = 0.0; // lambert components
sr = sg = sb = 0.0; // specular components


////////
// *** Scans for all the lights in the scene: ***
for(int i=0; i < sc->num_lights; i++)
////////
{

light = sc->lights[i];


// AMBIENT - quite simple: just sum up rgb color values
if( light->type == AMBIENT )
{
 lr += light->color.r;
 lg += light->color.g;
 lb += light->color.b;
 continue;
}

// DIRECTIONAL - only ('n' dot 'light_dir') weighting light's rgb
if( light->type == DIRECTIONAL )
{

if( g_cast_shadows )
{
 if( cast_shadows(sc, &light->dir_neg, sc->clip_far,  x,y,z) != NULL ) continue;
}

cos_alpha = prod_scalar( n, &light->dir_neg );

// if we are looking at the internal side of an opened surface
// (i.e.: if the angle between visual-ray and the normal is more than 90)
// we have to reverse the normal in order to achieve the right lambert component

if( cos_alpha < 0.0 ) continue;


if( prod_scalar( &ray->d, n )  >  0.0 ) 
{
 cos_alpha = -cos_alpha;
 inside = true;
}


 lr += light->color.r * cos_alpha;
 lg += light->color.g * cos_alpha;
 lb += light->color.b * cos_alpha;
 continue;
}




// calculate 'to light' vector:
to_l.x = light->pos.x - x;
to_l.y = light->pos.y - y;
to_l.z = light->pos.z - z;

//
// Check if we are outside the cube including light's action-radius
// (if yes skip to next light)
{
FTYPE rad = light->radius_falloff;

 if( to_l.x > rad   )  continue;
 if( to_l.y > rad   )  continue;
 if( to_l.z > rad   )  continue;

  rad = -rad;

 if( to_l.x < rad )  continue;
 if( to_l.y < rad )  continue;
 if( to_l.z < rad )  continue;
}


// Arrived Here? we are inside light's "cube-of-action": going on...

// From now on point-light distance is needed to compute
// attenuation factors and light clipping

// gets the length of 'to_l' vector (i.e: distance from the light)
dist2  =  POW2(to_l.x) + POW2(to_l.y) + POW2(to_l.z);


// checks if we are in the light's action-radius
if( dist2 > light->radius_falloff2 )  continue;




// for spot and lambert calculations
// we have to normalize the 'to light' vector
dist = SQRT( dist2 ); // normalize 'to light' vector:
t = 1.0 / dist;
 to_l.x *= t;
 to_l.y *= t;
 to_l.z *= t;


if(light->type==SPOT)
{
 // cosine of the angle between -1*'to_l' and light direction
 t1 = prod_scalar( &to_l, &light->dir_neg ); 

 // skip to next light if the point is beyond spot illumination angle
 if( t1 < light->angle_falloff_cos )  continue;
}


//
// 4 - Lambert component:
//

cos_alpha = prod_scalar( n, &to_l );

// if we are looking at the internal side of an opened surface
// (i.e.: if the angle between visual-ray and the normal is more than 90)
// we have reverse the normal to achieve the right lambert component
if( prod_scalar( &ray->d, n )  >  0.0 )
{
 cos_alpha = -cos_alpha;
 inside = true;
}

// skip to next light if the normal isn't facing the light
if( cos_alpha < 0 ) continue; 


// **********
//  Cast Shadows
// **********

// compute shadowing only if we are at the 1st recursion
if(recursion == g_num_recursions) 
if( g_cast_shadows ) // Shadow-casting enabled
{
  if( g_soft_shadows ) // Apply soft-shadows computation
  { 
   att = cast_soft_shadows( sc, &to_l,  x,y,z, light );

   if( att == 0.0 ) continue; // point completely obscured

   att *= light->multiplyer;
  }
   else // Classical sharp-shadows computations
   //if( g_t_min_ptr->shape_ptr->type!=PLANE )
  {
   if( cast_shadows( sc, &to_l, dist, x,y,z ) != NULL ) continue; // An obscurating object was hit!
   att = light->multiplyer;
  }

} 
 else // if shadow-casting is disabled directly set "att" to the light's multiplier val
  att = light->multiplyer;

  
//////////////////////////////////
// *** Attenuation factor: ***  //
//
// the 'attenuation factor' ("att") will tell us how much of the light's rgb
// we will take. We start from the light's multiplyer factor
// then 'att' is multiplied by each term contributing to attenuation
// In order the factors involved will be:
//
// 1 - light's multiplyer
// 2 - radial attenuation in the radial falloff region
// 3 - [SPOT] Spot's angular attenuation (if we are in the angular falloff region)
// 4 - Lambert's component ('normal' dot 'to_light')
//
//                             //
/////////////////////////////////
//
// 1 - Start!
// [ att = light->multiplyer ] already
 att *= cos_alpha; // multiply by "N" x "to_l"
//////



//
// 2 - Radial falloff attenuation (only if we are in the radial falloff region)
//
if( dist2 > light->radius2 )
{
//////
 att *= (light->radius_falloff2 - dist2) * light->radius_const;
//////
}




//
// 3 - Spot's angular falloff attenuation
//
if(light->type==SPOT)
{
// if we are into the spot's main illumination angle
// we won't have angular attenuation

 if( t1 < light->angle_cos )
 { 
	att *= ( t1 - light->angle_falloff_cos ) * light->angle_const;
    //att *= att; 	
 }

} // END 'SPOT'



/////
// 'att' is now ok!
/////


//-----------------------


// 
// *** Specular component: *** 
//
if(mat->shinyness > 0.0)
{
// compute reflected light-ray vector:

t = 2.0*(to_l.x*n->x + to_l.y*n->y + to_l.z*n->z);
 ref.x = -to_l.x + t*n->x;
 ref.y = -to_l.y + t*n->y;
 ref.z = -to_l.z + t*n->z;

// the vector from the point to the eye is -1*'ray->d'
// so compute the cosine of the angle between -1*'ray->d' and 'ref'

cos_beta = -prod_scalar( &ref, &ray->d );

 if(cos_beta > 0.0)
 {
 cos_beta = 1.0 - cos_beta;
 t = mat->shinyness;
 specular = -mat->shinyness_inv + (t+1.0) / ( t*(t*POW2(cos_beta) + 1.0) );
 
// specular=cos_beta*cos_beta;
// specular*=specular;


  // accumulate specular components:
  t = att * specular;
    sr += light->color.r * t;
    sg += light->color.g * t;
    sb += light->color.b * t;
  ////////////////////
 }


 
} 



////////////////////
// accumulate diffuse component:

  lr += light->color.r * att;
  lg += light->color.g * att;
  lb += light->color.b * att;


}  // END for( each light )

//-------------------------------


/*
r = r * lr + sr;
g = g * lg + sg;
b = b * lb + sb; 
*/
diff->r = lr;
diff->g = lg;
diff->b = lb;

spec->r = sr;
spec->g = sg;
spec->b = sb;



return inside;
}











__int32 raytrace(t_pixel *res, t_ray *ray0, t_scene *sc, int recursion )
{
FTYPE t_min;
FTYPE x, y, z;
t_material *mat;
FTYPE r, g, b;
t_vector3d n, dn;
t_point3d w, l;
t_ray *ray;
t_shape *sh_min;
t_intersection *t_min_ptr; // TODO: better t_min handling
bool inside;
int shape_id;





//=========================
if( recursion != g_num_recursions ) g_clip = false; else g_clip = true;

t_min_ptr = find_min_t( ray0, sc, NULL, sc->clip_far );

g_clip = false;
//=========================

//res->ray = ray0->d;


if( t_min_ptr == NULL ) 
{
// //res->values[0] = sc->fog_color.r;
// //res->values[1] = sc->fog_color.g;
// //res->values[2] = sc->fog_color.b;
 res->shape_id = 0;
 return RGBF2INT( sc->fog_color.r, sc->fog_color.g, sc->fog_color.b );
}


//
// 1: Get intersection coordinates:
//
sh_min = t_min_ptr->shape_ptr;

t_min = g_t_min;

ray = &sh_min->ray_inv;

shape_id = get_shape_id( sh_min, t_min_ptr );

// For some shapes intersection coordinates 
// are already computed during intersection
 if(sh_min->type == BOX || 
	  sh_min->type == TRIANGLE ||
		sh_min->type == TRILINEAR_CELL)
 {
  l.x = x = t_min_ptr->p.x;
  l.y = y = t_min_ptr->p.y;
  l.z = z = t_min_ptr->p.z;
 }
 else
{
 l.x = x = t_min * ray->d.x  +  ray->o.x;
 l.y = y = t_min * ray->d.y  +  ray->o.y;
 l.z = z = t_min * ray->d.z  +  ray->o.z;
}

ray = ray0;


//
// 2: Get the material
//
//res->mat = 
mat = sh_min->material;

res->texture = mat->texture;


//=====================





/*
if( TEX_IS_3D( mat->tex ) ) 
{
 //res->values[6] = FLOAT_2_FIX( x );
 //res->values[7] = FLOAT_2_FIX( y );
 //res->values[8] = FLOAT_2_FIX( z );
}
 else
 */
//if( mat->tex == TEX_IMAGE )
{
 // TODO: Texture Projection Function
 //get_texture_coord( &res->values[6], &res->values[7], sh_min, x, y, z);

}




//=====================

if( g_no_lighting )
{
 r = mat->diffuse.r;
 g = mat->diffuse.g;
 b = mat->diffuse.b;

 //res->values[0] = FLOAT_2_FIX( r );
 //res->values[1] = FLOAT_2_FIX( g );
 //res->values[2] = FLOAT_2_FIX( b );

 get_shape_normal( sh_min, &dn, x, y, z );


  // We have to transform the computed normal
  // and intersection coordinates in world coordinates
  // to compute lighting & reflections
  m_transform_normal( sh_min->m_inv, &dn, &n ); 

  if( prod_scalar( &ray->d, &n )  >  0.0 ) res->shape_id = - shape_id;

  else res->shape_id = shape_id;

  return RGBF2INT(r,g,b);
} // END if( g_no_lighting )



///////
// *** Beginning of light calculations: ***
//////

// Get the normal vector:
get_shape_normal( sh_min, &dn, x, y, z );

// We have to transform the computed normal
// and intersection coordinates in world coordinates
// to compute lighting & reflections in world space
m_transform_normal( sh_min->m_inv, &dn, &n );

// The transform could cause scalings along some axes
// so a normalization is required
norm_v( &n ); 

// Transform the local-coordinate intersection point
// in world coordinates
m_transform_point( sh_min->m, &l, &w );

// Store the world coordinates in x,y,z
x = w.x;
y = w.y;
z = w.z;



//FIX_TYPE *dest_vals = res->values;

// Create variables to store the diffusive & specular 
// components of lighting calculations
t_rgb l_diffuse, l_specular;

 inside = get_lighting( &l_diffuse, &l_specular,  sc, sh_min, ray, x, y, z, &n, recursion );


// We reverse the normal if we are looking at the "interior side" 
// (respect to the original normal)
/*
if( inside ) 
{
 n.x = -n.x;
 n.y = -n.y;
 n.z = -n.z;
}
*/

 //dest_vals[0] = FLOAT_2_FIX( l_diffuse.r );
 //dest_vals[1] = FLOAT_2_FIX( l_diffuse.g );
 //dest_vals[2] = FLOAT_2_FIX( l_diffuse.b );

 //dest_vals[3] = FLOAT_2_FIX256( l_specular.r );
 //dest_vals[4] = FLOAT_2_FIX256( l_specular.g );
 //dest_vals[5] = FLOAT_2_FIX256( l_specular.b );


if( g_use_texturing )
{
 FIX_TYPE ir, ig, ib, u, v;

 get_texture_coord( &u, &v, sh_min, l.x, l.y, l.z);

 if( g_bilinear_texturing ) 
	 texture256_get_pixel_B(&ir, &ig, &ib, mat->texture, u, v);
  else 
	 texture256_get_pixel(&ir, &ig, &ib, mat->texture, u, v);

  if( g_modulate_texturing )
  {
   r = mat->diffuse.r * (ir*(1.0/255.0));
   g = mat->diffuse.g * (ig*(1.0/255.0));
   b = mat->diffuse.b * (ib*(1.0/255.0));
	}
	 else
	{
   r = (ir*(1.0/255.0));
   g = (ig*(1.0/255.0));
   b = (ib*(1.0/255.0));
	}
}
else // no texturing
{
 r = mat->diffuse.r;
 g = mat->diffuse.g;
 b = mat->diffuse.b;
}






if(recursion<=0)
{
/*
 //res->values[0] = FLOAT_2_FIX( r );
 //res->values[1] = FLOAT_2_FIX( g );
 //res->values[2] = FLOAT_2_FIX( b );
*/
 if( inside ) res->shape_id = - shape_id;
  else res->shape_id = shape_id;

}

 else
// (recursion>0)
{
//
// REFLECTION:
//
if( mat->reflectance > 0.0 )
{
t_ray ref;
__int32 col;
t_pixel pix;

// compute reflected ray vector:
 // V = W - 2(W dot N)N

FTYPE t = -2.0*(ray->d.x*n.x + ray->d.y*n.y + ray->d.z*n.z);

 ref.d.x = ray->d.x  +  t * n.x;
 ref.d.y = ray->d.y  +  t * n.y;
 ref.d.z = ray->d.z  +  t * n.z;

norm_v(&ref.d);

ref.o.x = x + ref.d.x * EPSILON_INT;
ref.o.y = y + ref.d.y * EPSILON_INT;
ref.o.z = z + ref.d.z * EPSILON_INT;


col = raytrace( &pix, &ref, sc, recursion - 1 );

t = mat->reflectance;

 r = r*( (1.0-t) + (((col>>16)&0xFF) * (1.0/255.0)) * t);
 g = g*( (1.0-t) + (((col>>8 )&0xFF) * (1.0/255.0)) * t);
 b = b*( (1.0-t) + (((col    )&0xFF) * (1.0/255.0)) * t);


 //res->values[0] = FLOAT_2_FIX( r );
 //res->values[1] = FLOAT_2_FIX( g );
 //res->values[2] = FLOAT_2_FIX( b );



} // END reflection


//
// REFRACTION:
//
/*
if( mat->transparency > 0.0 && mat->ior != 0.0)
{
static FTYPE ior = mat->ior;
static FTYPE Ci  = -prod_scalar( &n, &ray->d );
static FTYPE Kn  = ior*Ci - SQRT( 1.0  +  ior*ior * (Ci*Ci - 1.0) );
static t_ray transmitted;


__int32 col;
t_pixel pix;



	transmitted.d.x  =  ior * ray->d.x  +  Kn * n.x;
	transmitted.d.y  =  ior * ray->d.y  +  Kn * n.y;
	transmitted.d.z  =  ior * ray->d.z  +  Kn * n.z;
	
norm_v(&transmitted.d);

transmitted.o.x = w.x + transmitted.d.x * EPSILON_INT;
transmitted.o.y = w.y + transmitted.d.y * EPSILON_INT;
transmitted.o.z = w.z + transmitted.d.z * EPSILON_INT;

col = raytrace( &pix, &transmitted, sc, recursion - 1 );

FTYPE t = mat->transparency;

 r += t*( ((col>>16)&0xFF) * (1.0/255.0) );
 g += t*( ((col>>8 )&0xFF) * (1.0/255.0) );
 b += t*( ((col    )&0xFF) * (1.0/255.0) );


} // END refraction
*/

} // END reflection/refraction




r = r * l_diffuse.r + l_specular.r;
g = g * l_diffuse.g + l_specular.g;
b = b * l_diffuse.b + l_specular.b;




//
// Compute fog shading:
//
if( g_use_fogging )
{


if(t_min > sc->fog_start ) 
{
 FTYPE t = (t_min - sc->fog_start);
 FTYPE t1;
 //t = ((my_noise3d(w.x+g_anim, w.y+g_anim*2.0, w.z, 1)*.5+.5)*t  );
 t = expf(-t*t*.02);// exp[f]() eats a lot of CPU cycles...
 t1 = 1.0 - t;

 r = r*t + sc->fog_color.r * t1;
 g = g*t + sc->fog_color.g * t1;
 b = b*t + sc->fog_color.b * t1;
}

} // END if( g_use_fogging )





// Clamp values:
 if(r > 1.0) r = 1.0;
 if(g > 1.0) g = 1.0;
 if(b > 1.0) b = 1.0;


/*
  function Expose(float light, float exposure)

    return ( 1 - exp(light * exposure) )  *  255

  end function


  main procedure:

    for every pixel on the screen

      calculate the amount of light entering the camera for this pixel, by Ray Tracing or whatever
      call this value L        

      PixelValue = Expose(L, K)

      plot this pixel with a brightness of PixelValue      

    end of loop

  end of main procedure
*/

 if( inside ) res->shape_id = - shape_id;

  else res->shape_id = shape_id;



return RGBF2INT(r,g,b)  |  ((inside ? shape_id|0x80 : shape_id)<<(16+8));


}








/*
 Renders 'scene' to the image buffer pointed by '*scr' (plus scr_w\h)
 using the 'cam' camera parameters for the viewpoint
*/

void render_scene0(t_camera *cam, t_scene *scene, t_pixel *pixel, int dst_w, int dst_h, int recursions)
{
int i, j;
__int32 *s;
FTYPE t;
t_ray ray;
t_point3d start;
t_vector3d delta_down, delta_left, ray_u, ray_v;
t_rgb col;


// Setup inverse transformation matrices

for(i=0; i < scene->num_objects; i++)
{
	m_invert( scene->shapes[i]->m, scene->shapes[i]->m_inv );
}


g_unknown = 0;




//
// Setup Projectors:
//

// compute the starting vector (from 'op' to 'top_left')
start = cam->top_left;
 start.x -= cam->op.x;
 start.y -= cam->op.y;
 start.z -= cam->op.z;



delta_left = cam->basis.u;
delta_down = cam->basis.v;

 prod_v_k( &delta_left, (cam->w) / (FTYPE)dst_w );

 prod_v_k( &delta_down, - (cam->h) / (FTYPE)dst_h );

 ray.o = cam->op;

 ray_v.x = start.x;
 ray_v.y = start.y;
 ray_v.z = start.z;

for(j=0; j < dst_h; j++)
{


  ray_u = ray_v;

  for(i=0; i < dst_w; i++)
  {

   ray.d = ray_u;

   norm_v( &ray.d ); // *** TODO: To optimize with Projectors' LUTs ***

//***===================***

     raytrace( pixel,  &ray, scene, recursions );

//***===================***


   pixel++;

   
   ray_u.x += delta_left.x;
   ray_u.y += delta_left.y;
   ray_u.z += delta_left.z;
  }

// shows the rendered scan-line:
//SDL_UpdateRect(g_SDL_buf, 0, j, dst_w, 1);

 // reaches the beginning of the new proj-plane line
  ray_v.x += delta_down.x;
  ray_v.y += delta_down.y;
  ray_v.z += delta_down.z;
}





#ifdef DEBUG_LOG
 printf("%i undefined shape-type raytrace request(s)\n", g_unknown);
#endif
}



#endif


