/*


  Ray 1.0 - Ray-Tracing: rendering core


*/


#ifndef _LIB_RENDER
#define _LIB_RENDER


#include "raytracer.h"


char g_mark[SCREEN_X*SCREEN_Y];
bool g_show_rays = false;



void get_clipping_rect( t_clipping_rect *r, t_shape *sh, t_camera *cam )
{


if( sh->radius <= 0.0 ) 
{
	r->visible = true;
	r->covers_all = true;
	return;
}

t_point3d p;
int x, y;
FTYPE kx, ky, kz;
char code;
int x_min, y_min, x_max, y_max;
bool inside;

FTYPE ax,ay,az,tx;
FTYPE bx,by,bz,ty;
FTYPE cx,cy,cz,tz;



if(sh->type == TRIANGLE)
{
t_point3d *v = ((t_triangle *)(sh->shape_data))->v;
FTYPE *m = sh->m;

FTYPE vx, vy, vz;

ax = m[0*4+0]; ay = m[0*4+1]; az = m[0*4+2];  tx = m[0*4+3];
bx = m[1*4+0]; by = m[1*4+1]; bz = m[1*4+2];  ty = m[1*4+3];
cx = m[2*4+0]; cy = m[2*4+1]; cz = m[2*4+2];  tz = m[2*4+3];

 for(int i=0; i<3; i++)
 {
  vx = v[i].x;
  vy = v[i].y;
  vz = v[i].z;

  p.x = ax*vx + ay*vy + az*vz  +  tx;
  p.y = bx*vx + by*vy + bz*vz  +  ty; 
  p.z = cx*vx + cy*vy + cz*vz  +  tz; 

  if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

  if(i)
  {
   if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
   if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;
  }
   else
   {
    x_min = x_max = x;
    y_min = y_max = y;
   }
 }

} // END triangle rect

else

{ // START box rect
FTYPE *m = sh->m;
t_point3d *c = &sh->corner;

ax = m[0*4+0]*c->x;
ay = m[0*4+1]*c->y;
az = m[0*4+2]*c->z;
tx = m[0*4+3];

bx = m[1*4+0]*c->x;
by = m[1*4+1]*c->y;
bz = m[1*4+2]*c->z;
ty = m[1*4+3];

cx = m[2*4+0]*c->x;
cy = m[2*4+1]*c->y;
cz = m[2*4+2]*c->z;
tz = m[2*4+3];

// Corner #1:
p.x = ax + ay + az  +  tx;
p.y = bx + by + bz  +  ty;
p.z = cx + cy + cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

  x_min = x_max = x;
  y_min = y_max = y;

// Corner #2:
p.x = ax + ay - az  +  tx;
p.y = bx + by - bz  +  ty;
p.z = cx + cy - cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;

// Corner #3:
p.x = ax - ay + az  +  tx;
p.y = bx - by + bz  +  ty;
p.z = cx - cy + cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;

// Corner #4:
p.x = ax - ay - az  +  tx;
p.y = bx - by - bz  +  ty;
p.z = cx - cy - cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;

///////////////////////////

// Corner #5:
p.x = -ax + ay + az  +  tx;
p.y = -bx + by + bz  +  ty;
p.z = -cx + cy + cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;

// Corner #6:
p.x = -ax + ay - az  +  tx;
p.y = -bx + by - bz  +  ty;
p.z = -cx + cy - cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;

// Corner #7:
p.x = -ax - ay + az  +  tx;
p.y = -bx - by + bz  +  ty;
p.z = -cx - cy + cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;

// Corner #8:
p.x = -ax - ay - az  +  tx;
p.y = -bx - by - bz  +  ty;
p.z = -cx - cy - cz  +  tz;

if( camera_project_point_screen( &p, cam,  &x, &y ) ) goto _LABEL_INSIDE;

	if( x < x_min ) x_min = x; else if( x > x_max ) x_max = x;
	if( y < y_min ) y_min = y; else if( y > y_max ) y_max = y;


}

/*****/
goto _LABEL_SKIP;
_LABEL_INSIDE:
	 r->covers_all = true;
   return;
_LABEL_SKIP:

	if( x_max < 0 || y_max < 0 || x_min >= SCREEN_X || y_min > SCREEN_Y ) r->visible = false;  
	 else 
		 {
		  char code = 0x00;

		  r->visible = true;

			if( x_min < 0 ) { x_min = 0; code|=0x01; }
			if( y_min < 0 ) { y_min = 0; code|=0x02; }
			if( x_max > SCREEN_X-1 ) { x_max = SCREEN_X-1; code|=0x04; }
			if( y_max > SCREEN_Y-1 ) { y_max = SCREEN_Y-1; code|=0x08; }

			if( code == 0x0F ) r->covers_all = true;
			  else r->covers_all = false;
		 }


 r->x_min = x_min;
 r->y_min = y_min;
 r->x_max = x_max;
 r->y_max = y_max;

}









void render_scene1(t_camera *cam, t_scene *scene, t_pixel *pixel, __int32 *vbuf, int dst_w, int dst_h, int recursions)
{
int i, j;
__int32 *s;
FTYPE t;
t_ray ray;
t_rgb col;
t_pixel *p = pixel;




// Setup inverse transformation matrices

for(i=0; i < scene->num_objects; i++)
{
	m_invert( scene->shapes[i]->m, scene->shapes[i]->m_inv );

	get_clipping_rect( &scene->shapes[i]->clipping_rect, scene->shapes[i], cam );
}




g_unknown = 0;


//
// Setup Projectors:
//
t_viewport *vp = &cam->viewport;
t_basis3d  *b  = &cam->basis;
t_vector3d proj;


ray.o = cam->op;

for(j=0; j <= SCREEN_Y; j+=QUAD_DIM)
{

  for(i=0; i <= SCREEN_X; i+=QUAD_DIM)
  {
   get_projector( &proj,  vp, i, j );
   transform_v( &ray.d,  &proj, b );
   raytrace( pixel, &ray, scene, recursions );
   pixel++;
  }

}

draw_quads( vbuf, &p[0], cam, scene );

#ifdef DEBUG_LOG
 printf("%i undefined shape-type raytrace request(s)\n", g_unknown);
#endif
}







bool interpolate2(__int32 *voff, int o1, int o2, int v)
{
__int32 c1 = *(voff + o1);
__int32 c2 = *(voff + o2);

// same shape_id (stored in the last higher 8 bits)
if( (c1 & 0xFF000000) != (c2 & 0xFF000000) ) return true;

int dr,dg,db;

dr = ((c1>>16)&0xFF) - ((c2>>16)&0xFF);
 if( ABS(dr) > v ) return true;
dg = ((c1>>8)&0xFF) - ((c2>>8)&0xFF);
 if( ABS(dg) > v ) return true;
db = ((c1)&0xFF) - ((c2)&0xFF);
 if( ABS(db) > v ) return true;
	
*voff = (( ((c1 & 0xFF00FF) + (c2 & 0xFF00FF)) >> 1)&0xFF00FF) | 
        (( ((c1 & 0x00FF00) + (c2 & 0x00FF00)) >> 1)&0x00FF00) | (c1 & 0xFF000000);


//					 *voff=0xFF0000;
	 return false;
}










void quad_render( __int32 *vbuf, 
								 t_camera *cam, t_scene *scene, int recursions, 
								 int quad_dim, int treshold )
{
__int32 *s;
t_vector3d proj;
t_ray ray;
int line_off, quad_dim_2;
t_viewport *vp;
t_basis3d  *b;
int i, j;
t_pixel pixel;


vp = &cam->viewport;
b  = &cam->basis;
ray.o = cam->op;

line_off = (quad_dim-1) * SCREEN_X;
quad_dim_2 = quad_dim / 2;



if(quad_dim==0)
{
s = vbuf;
for(j=0; j < SCREEN_Y; j+=QUAD_DIM)
{
g_j = j;
  for(i=0; i < SCREEN_X; i+=QUAD_DIM)
  {
	 g_i = i;
   get_projector( &proj,  vp, i, j );
   transform_v( &ray.d,  &proj, b );
   *s = raytrace( &pixel, &ray, scene, recursions );
	 if( g_show_rays ) g_mark[(s - vbuf)] = 1;

	 s += QUAD_DIM;
  }
	s += SCREEN_X * (QUAD_DIM-1);

}


s = vbuf + (SCREEN_Y-1) * SCREEN_X;
g_j = SCREEN_Y-1;
for(i=0; i < SCREEN_X; i++)
{
g_i = i;
   get_projector( &proj,  vp, i, SCREEN_Y-1 );
   transform_v( &ray.d,  &proj, b );
   *s = raytrace( &pixel, &ray, scene, recursions ); 
	if( g_show_rays ) g_mark[(s - vbuf)] = 1;

  s++;
}


s = vbuf + SCREEN_X-1;
g_i = SCREEN_X-1;
for(j=0; j < SCREEN_Y; j++)
{
g_j = j;

   get_projector( &proj,  vp, SCREEN_X-1, j );
   transform_v( &ray.d,  &proj, b );

    *s = raytrace( &pixel, &ray, scene, recursions ); 
		if( g_show_rays ) g_mark[(s - vbuf)] = 1;
  s += SCREEN_X;
}
return;
} // END if(quad_dim==0)


//
s = vbuf + quad_dim_2;

for(j=0; j < SCREEN_Y; j+=quad_dim)
{
	g_j = j;
  for(i=quad_dim_2; i < SCREEN_X; i+=quad_dim)
  {
		g_i = i;
    if( interpolate2(s, -quad_dim_2, quad_dim_2, treshold) )
    {
     get_projector( &proj,  vp, i, j );
     transform_v( &ray.d,  &proj, b );
     *s = raytrace( &pixel, &ray, scene, recursions); 
		 if( g_show_rays ) g_mark[(s - vbuf)] = 1;
		}
   s += quad_dim;
  }
	s += line_off;
}

//

int line_off_quad_dim_2 = quad_dim_2 * SCREEN_X;

s = vbuf  +  quad_dim_2 * SCREEN_X;

for(j=quad_dim_2; j < SCREEN_Y; j+=quad_dim)
{
g_j = j;
  for(i=0; i < SCREEN_X; i+=quad_dim)
  {
		g_i = i;
    if( interpolate2(s, -line_off_quad_dim_2, line_off_quad_dim_2, treshold) )
    {
     get_projector( &proj,  vp, i, j );
     transform_v( &ray.d,  &proj, b );
     *s = raytrace( &pixel, &ray, scene, recursions ); 
		 if( g_show_rays ) g_mark[(s - vbuf)] = 1;
		}
   s += quad_dim;
  }
	s += line_off;
}

//

s = vbuf  +  quad_dim_2 * (SCREEN_X) + quad_dim_2;
for(j=quad_dim_2; j < SCREEN_Y; j+=quad_dim)
{
	g_j = j;
  for(i=quad_dim_2; i < SCREEN_X; i+=quad_dim)
  {
    g_i = i;
    if( interpolate2(s, -quad_dim_2, quad_dim_2, treshold) )
    {
     get_projector( &proj,  vp, i, j );
     transform_v( &ray.d,  &proj, b );
     *s = raytrace( &pixel, &ray, scene, recursions ); 
		 if( g_show_rays ) g_mark[(s - vbuf)] = 1;
		}
   s += quad_dim;
  }
	s += line_off;
}

}






void render_scene2(t_camera *cam, t_scene *scene, t_pixel *pixel, __int32 *vbuf, int dst_w, int dst_h, int recursions)
{
int i, j;
__int32 *s;
FTYPE t;
t_ray ray;
t_rgb col;
t_pixel *p = pixel;


if( g_show_rays ) memset((void *)g_mark, 0, SCREEN_X*SCREEN_Y );


// Setup inverse transformation matrices

for(i=0; i < scene->num_objects; i++)
{
	m_invert( scene->shapes[i]->m, scene->shapes[i]->m_inv );

	get_clipping_rect( &scene->shapes[i]->clipping_rect, scene->shapes[i], cam );
}




g_unknown = 0;




t_viewport *vp = &cam->viewport;
t_basis3d  *b  = &cam->basis;
t_vector3d proj;
ray.o = cam->op;

quad_render( vbuf, cam, scene, recursions, 0  , 0 );

if( g_double_pixels )
{
	for(i=0; i<QUAD_SHIFT-1; i++)
	{
		handle_events();

		quad_render( vbuf, cam, scene, recursions, QUAD_DIM >> i, g_quality_parameter<<i );
	}

  s = vbuf;
	for(j=0; j<SCREEN_Y/2; j++, s+=SCREEN_X)
  for(i=0; i<SCREEN_X/2; i++, s+=2) *(s+1) = *(s+SCREEN_X) = *(s+SCREEN_X+1) = *s;
}
 else
{
 for(i=0; i<QUAD_SHIFT; i++)
 {
  handle_events();

  quad_render( vbuf, cam, scene, recursions, QUAD_DIM >> i, g_quality_parameter<<i );
 } 
}


if(g_show_rays)
{
 s = vbuf;
 for(i=0; i<SCREEN_X*SCREEN_Y; i++, s++)
 switch( g_mark[i] )
 {
  case 1: *s = 0x0000FF; break; 
  case 2: *s = 0x0F0FFF; break;
  case 3: *s = 0x3F3FFF; break;
  case 4: *s = 0x7F7FFF; break;
 }
}

}








#endif


