/*

  Ray 1.0 - Trilinear cell solvers

	(MuPAD)

*/

#ifndef _LIB_TRILINEAR_CELLS
#define _LIB_TRILINEAR_CELLS




/*
---------------
Trilinear cell:
---------------

      H_______G
     /|      /|
y   D-------C |
    | |     | |
^   | E_____|_F
| z |/      | /
|/  A_______B/
O--> x

A(0,0,0)


---------------------
MuPAD trilinear cell:
---------------------
x:=t*Dx+Ox;
y:=t*Dy+Oy;
z:=t*Dz+Oz;
eq:=a*(1-x)*(1-y)*(1-z) + b*x*(1-y)*(1-z) + c*x*y*(1-z) + d*(1-x)*y*(1-z) + e*(1-x)*(1-y)*z + f*x*(1-y)*z + g*x*y*z + h*(1-x)*y*z - ISO;
eqp:=poly(eq,[t]);
A:=nthcoeff(eqp,1);
B:=nthcoeff(eqp,2);
C:=nthcoeff(eqp,3);
Z:=tcoeff(eqp);

Mac:=generate::Macrofort:
Mac::init();
Mac::setOptimizedOption(TRUE):
Mac::openOutputFile("trilinear.f"):
Mac::genFor(["equal", [[c1, A],[c2, B],[c3, C],[c4, Z]]]):
Mac::closeOutputFile();
delete c1, c2, c3, c4;


----------------------
MuPAD trilinear cell gradient:
----------------------
eq:=a*(1-x)*(1-y)*(1-z) + b*x*(1-y)*(1-z) + c*x*y*(1-z) + d*(1-x)*y*(1-z) + e*(1-x)*(1-y)*z + f*x*(1-y)*z + g*x*y*z + h*(1-x)*y*z - ISO:

Gx:=diff(eq, x):
Gy:=diff(eq, y):
Gz:=diff(eq, z):

Mac:=generate::Macrofort:
Mac::init();
Mac::setOptimizedOption(TRUE):
Mac::openOutputFile("trilinear_G.f"):
Mac::genFor(["equal", [[gx, Gx],[gy, Gy],[gz, Gz]]]):
Mac::closeOutputFile();
delete gx, gy, gz;

	



// here some other look-ups 
// strange that MuPad did'nt optimized it out... anyway it's only Nuts.
// (maybe it's my fault in setting up some other parameters, who knows...) *8)

*/

//-----------------------------------------------------------

int solve_trilinear_cell(t_ray *ray, t_shape *sh)
{
/*
FTYPE a_b = ray->o.x*ray->d.x + ray->o.y*ray->d.y + ray->o.z*ray->d.z;
FTYPE a_a = POW2(ray->o.x) + POW2(ray->o.y) + POW2(ray->o.z);

// checks if the ray intersects the bounding sphere, if no return
if( POW2(a_b) < a_a - POW2(2.0) ) return 0;
*/


t_trilinear_cell *cell = (t_trilinear_cell *)sh->shape_data;
FTYPE Dx, Dy, Dz;
FTYPE Ox, Oy, Oz;

FTYPE A, B, C, D;

// pointer to cell values
FTYPE *v;
FTYPE a,b,c,d,e,f,g,h;

v = cell->val;
a = v[0];
b = v[1];
c = v[2];
d = v[3];
e = v[4];
f = v[5];
g = v[6];
h = v[7];

// ray parameters
Dx = ray->d.x;
Dy = ray->d.y;
Dz = ray->d.z;

Ox = ray->o.x;
Oy = ray->o.y;
Oz = ray->o.z;


// ** Nice Constants (doing it by hand would have been damaging!! W MuPAD!) ** 
/*
From here to the SolveCubic part there is a total of:
 72 MUL
 30 SUB  
 28 ADD 
 29 STO

 yeah.

*/

FTYPE t1,t2,t3,t4,t5,t6,
      t7,t8,t9,t10,t11,t12,
			t13,t14,t15,t16,t17,t18,
			t19,t20,t21,t22,t23;

 t1 = Dx*Oy;
 t2 = Dy*Ox;
 t3 = t1+t2;
 t4 = t3*Dz;
 t5 = 1-Oz;
 t6 = t5*Dx*Dy;
 t7 = 1-Oy;
 t8 = 1-Ox;
 t9 = t7*Dx*Dz;
 t10 = Dx*Oz;
 t11 = t7*Dx;
 t12 = t8*Dy;
 t14 = Dz*Ox;
 t15 = t10+t14;
 t16 = Dy*Oz;
 t17 = Dz*Oy;
 t18 = t16+t17;
 t19 = t11-t2;
 t20 = t12-t1;
 t21 = t8*Dz;
 t22 = t21-t10;
 t23 = -t12-t11;

 

B = a*(t6 - Dz*t23)+
    b*(-Dz*t19 - t6)+
    c*(t6 - t4)+
		d*(-Dz*t20 - t6)+
    e*(-Dy*t22 - t9)+
		f*(t9 - Dy*t15)+
    g*(t4 + Dx*t16)+
		h*(t8*Dy*Dz - Dx*t18);
      
 t13 = t17*Ox;
      
C = a*(t5*t23 - t7*t8*Dz) +
    b*(t5*t19 - t7*t14) +
		c*(t3*t5  - t13) +	
		d*(t5*t20 - t8*t17) +	
		e*(t7*t22 - t8*t16) +
		f*(t7*t15 - t16*Ox) +
		g*(t13    + t3*Oz) +
		h*(t8*t18 - t10*Oy);

D = ((g*Oz + c*t5)*Ox + (h*Oz + d*t5)*t8)*Oy +
		((b*t5 + f*Oz)*Ox + (a*t5 + e*Oz)*t8)*t7;
 
D -= cell->iso;


A = b+g+e+d-f-c-a-h;

double coeff[4], s[3];
int sols;

if(A!=0.0)  
{
 A *= Dx*Dy*Dz;

 coeff[3] = A;
 coeff[2] = B;
 coeff[1] = C;
 coeff[0] = D;

// solves a cubic in the form 'c[3]*x^3 + c[2]*x^2 + c[1]*x + c[0] = 0'
sols = solve_cubic( coeff, s );

while(sols--)  // filter negative intersections
   if( s[ sols ] > 0.0 ) add_intersection( s[ sols ], sh );
}
 else
{
A = 1.0/A;
 B *= A;
 C *= A;
 D *= A;

sols = solve_quadratic( B, C, D );

if(sols)     add_intersection( g_t1, sh );
if(sols==2)  add_intersection( g_t2, sh );
}


return 0;
}






void get_trilinear_cell_normal(t_vector3d *n, t_shape *sh, FTYPE x, FTYPE y, FTYPE z)
{
FTYPE *v = ((t_trilinear_cell *)sh->shape_data)->val;

FTYPE t1 = 1.0 - z;
FTYPE t2 = 1.0 - y;
FTYPE t3 = 1.0 - x;

 n->x = (v[6] - v[7])*y*z + (v[2]*y - v[3]*y)*t1 - ((v[4]*z + v[5]*z) - (v[0] + v[1])*t1)*t2;
 n->y = ((v[2] - v[1])*x + (v[3] - v[0])*t3)*t1 + ((v[6] - v[5])*x + (v[7] - v[4])*t3)*z;
 n->z = ((v[5] - v[1])*t2 + (v[6] - v[2])*y)*x + ((v[7] - v[3])*y + (v[4] - v[0])*t2)*t3;

norm_v( n );
}






#endif

